# Copyright (C) 2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Unit tests for the silly apps.
"""

from baldaquin.silly.common import SillyConfiguration, SillyPacket, SillyServer


def test_packet():
    """Test the basic packet structure.
    """
    _ = SillyPacket(0, 1, 13080, 120)


def test_server():
    """Test the event server.
    """
    server = SillyServer()
    for _ in range(5):
        data = server.next()
        _ = SillyPacket.unpack(data)


def test_configuration():
    """Test for the server configuration.
    """
    _ = SillyConfiguration()



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Basic package environment settings.
"""

import os
import pathlib
from typing import Tuple

from .logging_ import logger
from .typing_ import PathLike


def resolve(path: PathLike) -> pathlib.Path:
    """Resolve a given path, in the form of a pathlib.Path instance or a string.

    Path.resolve() returns an absolute, normalized path with all symbolic links
    resolved (if possible).

    Arguments
    ---------
    folder_path : Path instance
        The path to be resolved.

    Returns
    -------
    Path
        The resolved path.
    """
    return pathlib.Path(path).resolve()


def _mkdir(folder_path: PathLike) -> None:
    """Create a given folder if it does not exist.

    This is a small utility function to ensure that the relevant directories
    exist when needed at runtime.

    Arguments
    ---------
    folder_path : Path instance
        The path to the target folder.
    """
    folder_path = resolve(folder_path)
    if not folder_path.exists():
        logger.info(f"Creating folder {folder_path}...")
        pathlib.Path.mkdir(folder_path, parents=True)


# Basic package structure. Note we are using the typical src layout for the package.
BALDAQUIN_SOURCE = pathlib.Path(__file__).parent
BALDAQUIN_ROOT = BALDAQUIN_SOURCE.parent.parent
BALDAQUIN_GRAPHICS = BALDAQUIN_SOURCE / "graphics"
BALDAQUIN_ICONS = BALDAQUIN_GRAPHICS / "icons"
BALDAQUIN_SKINS = BALDAQUIN_GRAPHICS / "skins"

# System-wide character encoding.
BALDAQUIN_ENCODING = os.getenv("BALDAQUIN_ENCODING", "utf-8")

# User folders for data, scratch files, and configuration files. Note these can
# be overridden via environment variables.
_HOME = pathlib.Path.home()
BALDAQUIN_DATA = resolve(os.getenv("BALDAQUIN_DATA", _HOME / "baldaquindata"))
_mkdir(BALDAQUIN_DATA)

BALDAQUIN_SCRATCH = resolve(os.getenv("BALDAQUIN_SCRATCH", _HOME / ".baldaquinscratch"))
_mkdir(BALDAQUIN_SCRATCH)

BALDAQUIN_CONFIG = resolve(os.getenv("BALDAQUIN_CONFIG", _HOME / ".baldaquin"))
_mkdir(BALDAQUIN_CONFIG)


def config_folder_path(project_name: str) -> pathlib.Path:
    """Return the path to the configuration folder for a given project.

    Arguments
    ---------
    project_name : str
        The name of the project.

    Returns
    -------
    Path
        The path to the configuration folder.
    """
    return BALDAQUIN_CONFIG / project_name


def data_folder_path(project_name: str) -> pathlib.Path:
    """Return the path to the data folder for a given project.

    Arguments
    ---------
    project_name : str
        The name of the project.

    Returns
    -------
    Path
        The path to the data folder.
    """
    return BALDAQUIN_DATA / project_name


def setup_project(project_name: str) -> Tuple[pathlib.Path, pathlib.Path]:
    """Setup the folder structure for a given project.

    Arguments
    ---------
    project_name : str
        The name of the project.
    """
    config_folder = config_folder_path(project_name)
    app_config_folder = config_folder / "apps"
    data_folder = data_folder_path(project_name)
    folder_list = (config_folder, app_config_folder, data_folder)
    for folder_path in folder_list:
        _mkdir(folder_path)
    return folder_list



# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Time-related facilities.
"""

from __future__ import annotations

import calendar
import datetime
import time
from dataclasses import dataclass


class tzoffset(datetime.tzinfo):

    # pylint: disable=invalid-name

    """Minimal tzinfo class factory to create time-aware datetime objects.

    See https://docs.python.org/3/library/datetime.html#datetime.tzinfo
    for more details.

    Arguments
    ---------
    name : str
        The tzinfo object name.

    offset : float
        The UTC offset in seconds.
    """

    def __init__(self, name: str, offset: float) -> None:
        """Constructor.
        """
        self.name = name
        self.offset = datetime.timedelta(seconds=offset)

    def utcoffset(self, dt: datetime.datetime) -> float:
        """Overloaded method.
        """
        return self.offset

    def dst(self, dt: datetime.datetime) -> float:
        """Overloaded method.

        According to the documentation, this Return the daylight saving time (DST)
        adjustment, as a timedelta object or None if DST information isn’t known.
        """
        return None

    def tzname(self, dt: datetime.datetime) -> str:
        """Overloaded method.
        """
        return self.name


@dataclass
class Timestamp:

    """Small utility class to represent a timezone-aware timestamp.

    A Timestamp encodes three basic pieces of information:

    * a datetime object in the UTC time zone;
    * a datetime object in the local time zone;
    * a timestamp in seconds, relative to the origin of the parent timeline.

    Timestamp objects support subtraction aritmethics as a handy shortcut
    to calculate time differences.

    Timestamp objects also support serialization/deserialization through the
    :meth:`to_dict() <baldaquin.timeline.Timestamp.to_dict>` and
    :meth:`from_dict() <baldaquin.timeline.Timestamp.from_dict>`, which, in turn, use
    internally a string conversion to ISO format. (``datetime`` objects
    are implemented in C and therefore have no ``__dict__`` slot.)

    Arguments
    ---------
    utc_datetime : datetime.datetime
        The (timezone-aware) UTC datetime object corresponding to the timestamp.

    local_datetime : datetime.datetime
        The (timezone-aware) local datetime object corresponding to the timestamp.

    seconds : float
        The seconds elapsed since the origin of the parent timeline.
    """

    utc_datetime: datetime.datetime
    local_datetime: datetime.datetime
    seconds: float

    # These are the fields that need special handling when serializing/deserializing.
    _DATETIME_FIELDS = ("utc_datetime", "local_datetime")

    def to_dict(self) -> dict:
        """Serialization.
        """
        dict_ = {**self.__dict__}
        for key in self._DATETIME_FIELDS:
            dict_.update({key: dict_[key].isoformat()})
        return dict_

    @classmethod
    def from_dict(cls, **kwargs) -> Timestamp:
        """Deserialization.
        """
        for key in cls._DATETIME_FIELDS:
            kwargs.update({key: datetime.datetime.fromisoformat(kwargs[key])})
        return cls(**kwargs)

    def __sub__(self, other: Timestamp) -> float:
        """Overloaded operator to support timestamp subtraction.
        """
        return self.seconds - other.seconds

    def __str__(self) -> str:
        """String formatting.
        """
        return f"{self.local_datetime} ({self.seconds} s)"


class Timeline:

    # pylint: disable=too-few-public-methods

    """Class representing a continuos timeline referred to a fixed origin.

    Note that, by deafult, the origin of the Timeline is January 1, 1970 00:00:00 UTC,
    and the seconds field in the Timestamp objects that the Timeline returns
    when latched correspond to the standard POSIX time. Setting the origin to a
    different value allow, e.g., to emulate the mission elapsed time (MET)
    concept that is common in space missions.

    Arguments
    ---------
    origin : str
        A string representation (in ISO 8601 format) of the date and time
        corresponding to the origin of the timeline in UTC (without the
        traling +00:00).

        More specifically, according to the datetime documentation, we support
        strings in the format:

        .. code-block::

           YYYY-MM-DD[*HH[:MM[:SS[.fff[fff]]]]

        where * can match any single character.
    """

    _POSIX_ORIGIN = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)

    def __init__(self, origin: str = "1970-01-01") -> None:
        """Constructor.
        """
        self.origin = datetime.datetime.fromisoformat(f"{origin}")
        self.origin = self.origin.replace(tzinfo=datetime.timezone.utc)
        self._timestamp_offset = (self.origin - self._POSIX_ORIGIN).total_seconds()

    @staticmethod
    def _utc_offset() -> int:
        """Return the local UTC offset in s, considering the DST---note this has
        to be calculated every time the timeline is latched, as one doesn't know
        if a DST change has happened between two successive calls.

        See https://stackoverflow.com/questions/3168096 for more details on why
        this is a sensible way to calculate this.
        """
        return calendar.timegm(time.localtime()) - calendar.timegm(time.gmtime())

    def latch(self) -> Timestamp:
        """This is the workhorse function for keeping track of the time.

        This function latches the system time and creates a frozen Timestamp
        object that can be reused at later time.
        """
        # Retrieve the UTC date and time---this is preferred over datetime.utcnow(),
        # as the latter returns a naive datetime object, with tzinfo set to None.
        utc_datetime = datetime.datetime.now(datetime.timezone.utc)
        # Calculate the UTC offset.
        offset = self._utc_offset()
        # Add the offset to the UTC datetime and setup the tzinfo so that
        # the offset is included by default in the string representation.
        local_datetime = utc_datetime + datetime.timedelta(seconds=offset)
        local_datetime = local_datetime.replace(tzinfo=tzoffset("Local", offset))
        seconds = utc_datetime.timestamp() - self._timestamp_offset
        return Timestamp(utc_datetime, local_datetime, seconds)



# Copyright (C) 2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Plasduino sketch information.
"""

from pathlib import Path

from baldaquin.env import BALDAQUIN_SOURCE
from baldaquin.plasduino import PROJECT_NAME

PLASDUINO_SKETCH_ROOT = BALDAQUIN_SOURCE / PROJECT_NAME / "sketches"

# Dictionary holding the correspondence between the (sketch_id, sketch_version)
# tuple and the name of the actual sketch to be uploaded on the board.
# (This could have been handled in a more sensible way with a hash, but it's too late.)
_PLASDUINO_SKETCH_DICT = {
    (1, 3): "digital_timer.hex",
    (2, 3): "analog_sampling.hex"
}


def sketch_file_path(sketch_id: int, sketch_version: int) -> Path:
    """Return the full file path pointing to the compiled version of a given
    sketch.

    Arguments
    ---------
    sketch_id : int
        The identification number for the sketch.

    sketch_version : int
        The sketch version.

    Returns
    -------
    Path
        The full path to the binary file ready to be uploaded on the board.
    """
    try:
        file_name = _PLASDUINO_SKETCH_DICT[(sketch_id, sketch_version)]
    except KeyError as exc:
        raise RuntimeError(f"No information for sketch {sketch_id} version {sketch_version}") \
            from exc
    return PLASDUINO_SKETCH_ROOT / file_name



# Copyright (C) 2022--2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""User application framework.
"""

from typing import TYPE_CHECKING

from .__qt__ import QtCore
from .config import UserApplicationConfiguration
from .logging_ import logger

if TYPE_CHECKING:
    # We only need RunControl for type annotations, hence the if clause.
    from baldaquin.runctrl import RunControlBase


class UserApplicationBase:

    """Base class for user applications.
    """

    # pylint: disable=c-extension-no-member
    NAME = "User application"
    EVENT_HANDLER_CLASS = None
    CONFIGURATION_CLASS = None
    CONFIGURATION_FILE_PATH = None

    def __init__(self) -> None:
        """Constructor.
        """
        # pylint: disable=not-callable
        self.event_handler = self.EVENT_HANDLER_CLASS()
        # We should think about whether there is a more elegant way to do this.
        # Pass the user application to the child event handler? Use inheritance
        # rather than composition?
        self.event_handler.process_packet = self.process_packet
        self.configuration = self.CONFIGURATION_CLASS()
        if self.CONFIGURATION_FILE_PATH is not None:
            if self.CONFIGURATION_FILE_PATH.exists():
                self.configuration.update_from_file(self.CONFIGURATION_FILE_PATH)
            else:
                self.configuration.save(self.CONFIGURATION_FILE_PATH)

    def apply_configuration(self, configuration: UserApplicationConfiguration):
        """Set the configuration for the user application.
        """
        self.configuration = configuration
        if self.CONFIGURATION_FILE_PATH is not None:
            self.configuration.save(self.CONFIGURATION_FILE_PATH)
        self.configure()

    def configure(self):
        """Apply a given configuration to the user application.
        """
        raise NotImplementedError

    def setup(self) -> None:
        """Function called when the run control transitions from RESET to STOPPED.
        """
        logger.info(f"{self.__class__.__name__}.setup(): nothing to do...")

    def teardown(self) -> None:
        """Function called when the run control transitions from STOPPED to RESET.
        """
        logger.info(f"{self.__class__.__name__}.teardown(): nothing to do...")

    def start_run(self) -> None:
        """Start the event handler.
        """
        logger.info(f"Starting {self.NAME} user application...")
        self.event_handler.reset_statistics()
        QtCore.QThreadPool.globalInstance().start(self.event_handler)

    def stop_run(self) -> None:
        """Stop the event handler.
        """
        logger.info(f"Stopping {self.NAME} user application...")
        self.event_handler.stop()
        QtCore.QThreadPool.globalInstance().waitForDone()
        self.event_handler.flush_buffer()

    def pause(self) -> None:
        """Pause the event handler.
        """
        logger.info(f"Pausing {self.NAME} user application...")
        self.event_handler.stop()
        QtCore.QThreadPool.globalInstance().waitForDone()
        self.event_handler.flush_buffer()

    def resume(self) -> None:
        """Resume the event handler.
        """
        logger.info(f"Resuming {self.NAME} user application...")
        QtCore.QThreadPool.globalInstance().start(self.event_handler)

    def stop(self) -> None:
        """Stop the event handler.
        """
        self.stop_run()

    def pre_start(self, run_control: "RunControlBase") -> None:
        """Hook that subclasses can use to perform any operation that needs to
        be done right before the application is stared (e.g., adding a custom
        sink to the underlying packet buffer.)
        """

    def post_stop(self, run_control: "RunControlBase") -> None:
        """Hook that subclasses can use to post-process the data collected in
        the run.
        """

    def process_packet(self, packet_data):
        """Optional hook for a user application to do something with the event data.
        """



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""baldaquin command-line interface.
"""

from __future__ import annotations

import argparse
import importlib
import importlib.util
import sys

import aptapy.plotting

from baldaquin import __version__, arduino_, serial_
from baldaquin.env import BALDAQUIN_DATA, BALDAQUIN_SOURCE
from baldaquin.logging_ import logger

# List of defaults projects shipped with the package.
_DEFAULT_PROJECTS = ("plasduino", "silly")


def start_message() -> None:
    """Print the start message.
    """
    msg = f"""
    This is baldaquin version {__version__}.

    Copyright (C) 2022--2025, the baldaquin team.

    baldaquin comes with ABSOLUTELY NO WARRANTY.
    This is free software, and you are welcome to redistribute it under certain
    conditions. See the LICENSE file for details.

    Visit https://github.com/lucabaldini/baldaquin for more information.
    """
    print(msg)


class _Formatter(argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter):

    """Do nothing class combining our favorite formatting for the
    command-line options, i.e., the newlines in the descriptions are
    preserved and, at the same time, the argument defaults are printed
    out when the --help options is passed.

    The inspiration for this is coming from one of the comments in
    https://stackoverflow.com/questions/3853722
    """


class MainArgumentParser(argparse.ArgumentParser):

    """Application-wide argument parser.
    """

    _DESCRIPTION = None
    _EPILOG = None
    _FORMATTER_CLASS = _Formatter

    def __init__(self) -> None:
        """Overloaded method.
        """
        super().__init__(description=self._DESCRIPTION, epilog=self._EPILOG,
                         formatter_class=self._FORMATTER_CLASS)
        subparsers = self.add_subparsers(required=True, help="sub-command help")
        # See https://stackoverflow.com/questions/8757338/
        subparsers._parser_class = argparse.ArgumentParser

        # Start a given application.
        start_app = subparsers.add_parser("start-app",
            help="start a baldaquin application",
            formatter_class=self._FORMATTER_CLASS)
        start_app.add_argument("app_name", choices=self.list_apps())
        start_app.add_argument("--mplstyle", default="aptapy",
                               help="the matplotlib stylesheet to apply")
        start_app.set_defaults(func=self.start_app)

        # Simply list the COM ports.
        list_com_ports = subparsers.add_parser("list-com-ports",
            help="list the available COM ports",
            formatter_class=self._FORMATTER_CLASS)
        list_com_ports.set_defaults(func=serial_.list_com_ports)

        # Arduino autodetect.
        arduino_autodetect = subparsers.add_parser("arduino-autodetect",
            help="autodetect arduino boards attached to the COM ports",
            formatter_class=self._FORMATTER_CLASS)
        arduino_autodetect.set_defaults(func=arduino_.autodetect_arduino_boards)

        # Arduino upload.
        arduino_compile = subparsers.add_parser("arduino-compile",
            help="compile a sketch for a given arduino board",
            formatter_class=self._FORMATTER_CLASS)
        arduino_compile.add_argument("file_path",
            help="the path to the sketch source file")
        arduino_compile.add_argument("--output_dir", default=BALDAQUIN_DATA)
        arduino_compile.add_argument("--board_designator", default="uno")
        arduino_compile.set_defaults(func=arduino_.compile_sketch)

        # Arduino upload.
        arduino_upload = subparsers.add_parser("arduino-upload",
            help="upload a sketch to an arduino board",
            formatter_class=self._FORMATTER_CLASS)
        arduino_upload.add_argument("file_path",
            help="the path to the compiled sketch file")
        arduino_upload.add_argument("--board-designator", default="uno")
        arduino_upload.set_defaults(func=arduino_.upload_sketch)

    @staticmethod
    def list_apps() -> list[str]:
        """List all the available applications.

        This is looping over all the registered project folders and searching
        for Python modules. Note that ``pathlib.Path.iterdir()`` returns a list
        in no particular order, and we do sort the list in place for convenience.
        """
        apps = []
        for project in _DEFAULT_PROJECTS:
            folder_path = BALDAQUIN_SOURCE / project / "apps"
            apps += [_path.stem for _path in folder_path.iterdir() if _path.suffix == ".py"]
        apps.sort()
        return apps

    @staticmethod
    def start_app(app_name: str, **kwargs) -> None:
        """Start a given application.

        Arguments
        ---------
        app_name : str
            The application name.
        """
        try:
            aptapy.plotting.apply_stylesheet(kwargs.get("mplstyle"))
        except Exception as exc:
            logger.warning(f"Could not apply the requested stylesheet: {exc}")
        logger.info("Starting application...")
        # Loop over the project folders and search for a Python module matching the
        # target application name.
        for project in _DEFAULT_PROJECTS:
            folder_path = BALDAQUIN_SOURCE / project / "apps"
            file_path = folder_path / f"{app_name}.py"
            if file_path.exists():
                logger.debug(f"Adding {file_path} to sys.path...")
                sys.path.append(str(folder_path))
                break
        # At this point we do want to import the module without generating the
        # bytecode---we cannot assume the thing is in a folder with write priviledges.
        sys.dont_write_bytecode = True
        module = importlib.import_module(app_name)
        sys.dont_write_bytecode = False
        # And, finally, we call the application entry point.
        module.main()

    def run_command(self) -> None:
        """Run the actual command tied to the specific options.
        """
        kwargs = vars(self.parse_args())
        command = kwargs.pop("func")
        command(**kwargs)


def main() -> None:
    """Main entry point.
    """
    start_message()
    MainArgumentParser().run_command()


if __name__ == "__main__":
    main()



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pathlib

import pytest


@pytest.fixture(scope="session")
def test_data_dir() -> pathlib.Path:
    """Return the path to the test data directory.
    """
    return pathlib.Path(__file__).parent / "data"


@pytest.fixture
def test_data_path(test_data_dir: pathlib.Path):
    """Return a function to get a specific test data file path.
    """
    # pylint: disable=redefined-outer-name
    def _get(*parts: str) -> pathlib.Path:
        p = test_data_dir.joinpath(*parts)
        assert p.exists(), f"Missing test data: {p}"
        return p
    return _get



# Copyright (C) 2024--25 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Plasduino temperature monitor application.
"""

from pathlib import Path

from baldaquin.buf import WriteMode
from baldaquin.egu import ThermistorConversion
from baldaquin.env import BALDAQUIN_ENCODING
from baldaquin.gui import bootstrap_window
from baldaquin.logging_ import logger
from baldaquin.pkt import AbstractPacket, PacketFile, packetclass
from baldaquin.plasduino import PLASDUINO_APP_CONFIG, PLASDUINO_SENSORS
from baldaquin.plasduino.common import (
    PlasduinoAnalogConfiguration,
    PlasduinoAnalogEventHandler,
    PlasduinoAnalogUserApplicationBase,
    PlasduinoMainWindow,
    PlasduinoRunControl,
)
from baldaquin.plasduino.protocol import AnalogReadout
from baldaquin.plasduino.shields import Lab1
from baldaquin.runctrl import RunControlBase


class AppMainWindow(PlasduinoMainWindow):

    pass


@packetclass
class TemperatureReadout(AnalogReadout):

    """Specialized class inheriting from ``AnalogReadout`` describing a temperature
    readout---this is essentially adding the conversion between ADC counts and
    temperature on top of the basic functions.

    We have decided to go this route for two reasons:

    * it makes it easy to guarantee that the conversion is performed once and
      forever when the packet object is created;
    * it allows to easily implement the text conversion.
    """

    _CONVERSION_FILE_PATH = PLASDUINO_SENSORS / "NXFT15XH103FA2B.dat"
    _ADC_NUM_BITS = 10
    _CONVERSION_COLS = (0, 2)
    _CONVERTER = ThermistorConversion.from_file(_CONVERSION_FILE_PATH, Lab1.SHUNT_RESISTANCE,
                                                _ADC_NUM_BITS, *_CONVERSION_COLS)

    OUTPUT_HEADERS = ("Pin number", "Time [s]", "Temperature [deg C]")
    OUTPUT_ATTRIBUTES = ("pin_number", "seconds", "temperature")
    OUTPUT_FMTS = ("%d", "%.3f", "%.2f")

    def __post_init__(self) -> None:
        """Post initialization.
        """
        AnalogReadout.__post_init__(self)
        self.temperature = self._CONVERTER(self.adc_value)


class TemperatureMonitor(PlasduinoAnalogUserApplicationBase):

    """Simplest possible user application for testing purposes.
    """

    NAME = "Temperature Monitor"
    CONFIGURATION_CLASS = PlasduinoAnalogConfiguration
    CONFIGURATION_FILE_PATH = PLASDUINO_APP_CONFIG / "plasduino_tempmonitor.cfg"
    EVENT_HANDLER_CLASS = PlasduinoAnalogEventHandler
    _PINS = Lab1.TEMPMON_PINS
    _LABEL = "Temperature [deg C]"
    _SAMPLING_INTERVAL = 500

    def process_packet(self, packet_data: bytes) -> AbstractPacket:
        """Overloaded method.
        """
        readout = TemperatureReadout.unpack(packet_data)
        x, y = readout.seconds, readout.temperature
        self.strip_chart_dict[readout.pin_number].put(x, y)
        return readout

    def post_process_file(self, file_path: Path, delimiter: str = "   ") -> None:
        """Post-process a binary data file to produce a text output file.
        """
        logger.info(f"Post-processing file {file_path}...")
        output_file_path = file_path.with_name(file_path.stem + "_proc.txt")
        header = AbstractPacket.text_header(prefix="# ", creator=self.NAME)
        fields = []
        for pin in self._PINS:
            fields += [f"Time{pin}", f"Temp{pin}"]
        header = f"{header}# {delimiter.join(fields)}\n"
        with PacketFile(TemperatureReadout).open(file_path) as input_file, \
             open(output_file_path, "w", encoding=BALDAQUIN_ENCODING) as output_file:
            row = []
            output_file.write(header)
            for readout in input_file:
                row += [f"{readout.seconds:.3f}", f"{ readout.temperature:.2f}"]
                if readout.pin_number == self._PINS[-1]:
                    output_file.write(f"{delimiter.join(row)}\n")
                    row = []
        logger.info(f"Post-processed data written to {output_file_path}")

    def pre_start(self, run_control: RunControlBase) -> None:
        """Overloaded method.
        """
        # If we are starting a run after the completion of a previous one, deactivate
        # the previous interactive cursor and delete the corresponding reference.
        self.deactivate_cursors()
        # And create the sink for the output text file.
        file_path = Path(f"{run_control.output_file_path_base()}_data.csv")
        self.event_handler.add_custom_sink(file_path, WriteMode.TEXT, TemperatureReadout.to_text,
                                           TemperatureReadout.text_header(creator=self.NAME))

    def post_stop(self, run_control: RunControlBase) -> None:
        """Overloaded method.

        This is where we re-read all the data from disk to populate the complete
        strip charts, and then enable the vertical cursor.
        """
        self.post_process_file(run_control.data_file_path())
        logger.debug("Clearing strip charts...")
        # First thing first, set to None the maximum length for all the strip charts
        # to allow unlimited deque size. (Note this creates two new deques under the
        # hood, so we don't need to clear the strip charts explicitly. Also note
        # that the proper maximum length will be re-applied in the configure() slot,
        # based on the value from the GUI.)
        for chart in self.strip_chart_dict.values():
            chart.set_max_length(None)
        # Read all the data from disk and rebuild the entire strip charts.
        logger.debug("Re-building strip charts from disk...")
        with PacketFile(TemperatureReadout).open(run_control.data_file_path()) as input_file:
            for readout in input_file:
                x, y = readout.seconds, readout.temperature
                self.strip_chart_dict[readout.pin_number].put(x, y)
        # Plot the strip charts and activate the vertical cursor.
        self.activate_cursors()


def main() -> None:
    """Main entry point.
    """
    bootstrap_window(AppMainWindow, PlasduinoRunControl(), TemperatureMonitor())


if __name__ == "__main__":
    main()



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for pkt.py
"""

import pytest

from baldaquin.env import BALDAQUIN_SCRATCH
from baldaquin.pkt import (
    AbstractPacket,
    FieldMismatchError,
    FixedSizePacketBase,
    Format,
    Layout,
    PacketFile,
    PacketStatistics,
    packetclass,
)


@packetclass
class Readout(FixedSizePacketBase):

    """Plausible data structure for testing purposes.
    """

    layout = Layout.BIG_ENDIAN
    header: Format.UNSIGNED_CHAR = 0xaa
    milliseconds: Format.UNSIGNED_LONG
    adc_value: Format.UNSIGNED_SHORT

    def __post_init__(self) -> None:
        self.seconds = self.milliseconds / 1000.


def test_format():
    """Test the check for the packet layout and format characters.
    """
    # pylint: disable=function-redefined, unused-variable
    with pytest.raises(ValueError):
        @packetclass
        class Packet(FixedSizePacketBase):
            layout = "W"

    with pytest.raises(ValueError):
        @packetclass
        class Packet(FixedSizePacketBase):  # noqa: F811
            trigger_id: "W"  # noqa: F821

    with pytest.raises(FieldMismatchError):
        _ = Readout(0, 0, 0)


def test_readout():
    """Test a sensible packet structure.
    """
    # pylint: disable=protected-access
    # Test the class variables.
    assert Readout._fields == ("header", "milliseconds", "adc_value")
    assert Readout._format == ">BLH"
    assert Readout.size == 7

    # Create a class instance.
    packet = Readout(0xaa, 100, 127)
    assert isinstance(packet, AbstractPacket)
    # Test the post-initialization.
    assert packet.seconds == packet.milliseconds / 1000.
    # Make sure that pack/unpack do roundtrip.
    twin = Readout.unpack(packet.pack())
    for val1, val2 in zip(packet, twin):
        assert val1 == val2
    # Make sure that the packet fields cannot be modified.
    with pytest.raises(AttributeError):
        packet.header = 3


def test_readout_subclass():
    """Test subclassing a concrete FixedSizePacketBase subclass.
    """

    @packetclass
    class SpecialReadout(Readout):

        board_number: Format.UNSIGNED_SHORT

    readout = SpecialReadout(0xaa, 100, 127, 3)
    assert readout.milliseconds == 100
    assert readout.adc_value == 127
    assert readout.board_number == 3


def test_text():
    """Test the join_attributes() method.
    """
    # pylint: disable=protected-access
    attrs = ("seconds", "adc_value")
    fmts = ("%.6f", "%d")
    packet = Readout(0xaa, 100, 127)
    assert packet._format_attributes(attrs, fmts) == ("0.100000", "127")
    assert packet._text(attrs, fmts, ", ") == "0.100000, 127\n"


def test_repr():
    """Test the terminal formatting helper function.
    """
    # pylint: disable=protected-access
    attrs = ("seconds", "adc_value")
    fmts = ("%.6f", "%d")
    packet = Readout(0xaa, 100, 127)
    assert packet._repr(attrs) == "Readout(seconds=0.1, adc_value=127)"
    assert packet._repr(attrs, fmts) == "Readout(seconds=0.100000, adc_value=127)"


def test_docs():
    """Small convenience function for the class docs---we copy/paste from here.
    """
    # pylint: disable=function-redefined
    @packetclass
    class Trigger(FixedSizePacketBase):

        layout = Layout.BIG_ENDIAN

        header: Format.UNSIGNED_CHAR = 0xff
        pin_number: Format.UNSIGNED_CHAR
        timestamp: Format.UNSIGNED_LONG_LONG

    packet = Trigger(0xff, 1, 15426782)
    assert len(packet) == 10
    assert isinstance(packet, AbstractPacket)

    packet = Trigger.unpack(b"\xff\x01\x00\x00\x00\x00\x00\xebd\xde")

    with pytest.raises(AttributeError):
        packet.pin_number = 0

    @packetclass
    class Trigger(FixedSizePacketBase):

        layout = Layout.BIG_ENDIAN

        header: Format.UNSIGNED_CHAR = 0xff
        pin_number: Format.UNSIGNED_CHAR
        microseconds: Format.UNSIGNED_LONG_LONG

        def __post_init__(self):
            self.seconds = self.microseconds / 1000000

    packet = Trigger(0xff, 1, 15426782)
    assert packet.seconds == 15.426782


def test_binary_io(num_packets: int = 10):
    """Write to and read from file in binary format.
    """
    file_path = BALDAQUIN_SCRATCH / "test_pkt.dat"
    with open(file_path, "wb") as output_file:
        for i in range(num_packets):
            output_file.write(Readout(0xaa, 1 * 1000, i + 100).data)
    with PacketFile(Readout).open(file_path) as input_file:
        for _ in input_file:
            pass
    with PacketFile(Readout).open(file_path) as input_file:
        packets = input_file.read_all()
        assert len(packets) == num_packets


def test_packets_statistics():
    """Small test for the PacketStatistics class.
    """
    stats = PacketStatistics()
    stats.update(3, 3, 10)



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""xnucleo environmental monitor application.
"""


import datetime
import time
from dataclasses import dataclass
from pathlib import Path

from aptapy.strip import EpochStripChart

from baldaquin import xnucleo
from baldaquin.__qt__ import QtWidgets
from baldaquin.app import UserApplicationBase
from baldaquin.arduino_ import ArduinoEventHandler
from baldaquin.buf import WriteMode
from baldaquin.config import UserApplicationConfiguration
from baldaquin.gui import MainWindow, SimpleControlBar, bootstrap_window
from baldaquin.logging_ import logger
from baldaquin.pkt import AbstractPacket
from baldaquin.runctrl import RunControlBase


@dataclass
class MonitorReadout(AbstractPacket):

    """Class describing a full readout for the xnucleo monitor application.
    """

    seconds: float
    humidity: float
    temperature1: float
    pressure: float
    temperature2: float
    adc1: int
    adc2: int
    _data: bytes = None

    OUTPUT_HEADERS = ("Time [s]", "Temperature 1 [deg C]", "Temperature 2 [deg C]",
                      "Relative humidity [%]", "Pressure [mbar]",
                      "Channel 1 [ADC counts]", "Channel 2 [ADC counts]")
    OUTPUT_ATTRIBUTES = ("seconds", "temperature1", "temperature2", "humidity",
                         "pressure", "adc1", "adc2")
    OUTPUT_FMTS = ("%.3f", "%.2f", "%.2f", "%.2f", "%.2f", "%d", "%d")

    @property
    def data(self) -> bytes:
        """Return the packet binary data.
        """
        return self._data

    @property
    def fields(self) -> tuple:
        """Return the packet fields.
        """
        return

    def __len__(self) -> int:
        """Return the length of the binary data in bytes.
        """
        return len(self._data)

    def __iter__(self):
        """Iterate over the field values.
        """
        raise NotImplementedError

    def pack(self) -> bytes:
        """Pack the field values into the corresponding binary data.
        """
        raise NotImplementedError

    @classmethod
    def unpack(cls, data: bytes):
        """Unpack the binary data into the corresponding field values.
        """
        fields = data.unpack(float, float, float, float, float, int, int)
        return cls(*fields, data)

    @classmethod
    def text_header(cls, prefix: str = "#", creator: str = None) -> str:
        """Return the text header.
        """
        headers = ", ".join(map(str, cls.OUTPUT_HEADERS))
        return f"{AbstractPacket.text_header(prefix, creator)}{prefix}{headers}\n"

    def to_text(self, separator: str = ",") -> str:
        """Overloaded method.
        """
        return self._text(self.OUTPUT_ATTRIBUTES, self.OUTPUT_FMTS, separator)


class MonitorWindow(MainWindow):

    """Application graphical user interface.
    """

    _PROJECT_NAME = xnucleo.PROJECT_NAME
    _CONTROL_BAR_CLASS = SimpleControlBar

    def __init__(self, parent: QtWidgets.QWidget = None) -> None:
        """Constructor.
        """
        super().__init__()
        self.hide_reset_button()
        self.temperature_tab = self.add_plot_canvas_tab("Temperature")
        self.humidity_tab = self.add_plot_canvas_tab("Humidity")
        self.pressure_tab = self.add_plot_canvas_tab("Pressure")
        self.analog_tab = self.add_plot_canvas_tab("Analog inputs")

    def setup_user_application(self, user_application):
        """Overloaded method.
        """
        super().setup_user_application(user_application)
        self.temperature_tab.register(user_application.temperature1_strip_chart,
                                      user_application.temperature2_strip_chart)
        self.humidity_tab.register(user_application.humidity_strip_chart)
        self.pressure_tab.register(user_application.pressure_strip_chart)
        self.analog_tab.register(user_application.adc1_strip_chart,
                                 user_application.adc2_strip_chart)


class MonitorConfiguration(UserApplicationConfiguration):

    _PARAMETER_SPECS = (
        ("sampling_interval", float, 2., "Sampling interval [s]",
            dict(min=1., max=1000.0)),
        ("strip_chart_max_length", int, 200, "Strip chart maximum length",
            dict(min=10, max=1000000)),
    )


class MonitorEventHandler(ArduinoEventHandler):

    """Base class for all the xnucleo event handlers.
    """

    _DEFAULT_SAMPLING_INTERVAL = 1.0

    def __init__(self) -> None:
        """Constructor.
        """
        super().__init__()
        self._sampling_interval = self._DEFAULT_SAMPLING_INTERVAL

    def set_sampling_interval(self, interval: float) -> None:
        """Set the sampling interval.
        """
        logger.info(f"Setting {self.__class__.__name__} sampling interval to {interval} s...")
        self._sampling_interval = interval

    def read_packet(self) -> bytes:
        """Basic function fetching a single readout from the xnucleo board attached
        to the arduino.

        This looks somewhat more convoluted than it might be for the simple reason
        that we are trying to stick to the basic baldaquin protocol, where data
        are fetched from the board interfaced to the host PC, and the latter is just
        waiting. Here, instead, we are sending arduino a byte on the serial port
        to trigger a readout, we are latching the timestamp on the host PC, we are
        reading the data from the serial port, and we are assembling everything
        together in a single bytes object that is then written to disk in binary format.
        """
        # Trigger a readout on the arduino board.
        self.serial_interface.pack_and_write(1, "B")
        # Latch the timestamp (seconds since the epoch, UTC) on the host PC.
        timestamp = datetime.datetime.now(datetime.timezone.utc).timestamp()
        # Wait...
        time.sleep(self._sampling_interval)
        # Read the data from the serial port...
        data = self.serial_interface.read_text_line()
        # ... and prepend the timestamp.
        data.prepend(f"{timestamp:.3f}")
        return data


class MonitorRunControl(RunControlBase):

    """Specialized xnucleo run control.
    """

    _PROJECT_NAME = xnucleo.PROJECT_NAME


class Monitor(UserApplicationBase):

    """Simplest possible user application for testing purposes.
    """

    NAME = "Generic Monitor"
    CONFIGURATION_CLASS = MonitorConfiguration
    CONFIGURATION_FILE_PATH = xnucleo.XNUCLEO_APP_CONFIG / "xnucleo_monitor.cfg"
    EVENT_HANDLER_CLASS = MonitorEventHandler
    SKETCH_NAME = "xnucleo_monitor"
    SKETCH_VERSION = 3

    def __init__(self) -> None:
        """Overloaded Constructor.
        """
        super().__init__()
        kwargs = dict(ylabel="Temperature [deg C]")
        self.temperature1_strip_chart = EpochStripChart(label="Temperature 1", **kwargs)
        self.temperature2_strip_chart = EpochStripChart(label="Temperature 2", **kwargs)
        kwargs = dict(ylabel="Humidity [%]")
        self.humidity_strip_chart = EpochStripChart(**kwargs)
        kwargs = dict(ylabel="Pressure [mbar]")
        self.pressure_strip_chart = EpochStripChart(**kwargs)
        kwargs = dict(ylabel="Value [ADC counts]")
        self.adc1_strip_chart = EpochStripChart(label="Channel 1", **kwargs)
        self.adc2_strip_chart = EpochStripChart(label="Channel 2", **kwargs)
        self._strip_charts = (self.temperature1_strip_chart, self.temperature2_strip_chart,
                              self.humidity_strip_chart, self.pressure_strip_chart,
                              self.adc1_strip_chart, self.adc2_strip_chart)

    def configure(self) -> None:
        """Overloaded method.
        """
        config = self.configuration.application_section()
        self.event_handler.set_sampling_interval(config.value("sampling_interval"))
        for chart in self._strip_charts:
            chart.set_max_length(config.value("strip_chart_max_length"))

    def setup(self) -> None:
        """Overloaded method (RESET -> STOPPED).
        """
        self.event_handler.open_serial_interface()
        args = self.SKETCH_NAME, self.SKETCH_VERSION, xnucleo.XNUCLEO_SKETCHES
        self.event_handler.serial_interface.handshake(*args)

    def teardown(self) -> None:
        """Overloaded method (STOPPED -> RESET).
        """
        self.event_handler.close_serial_interface()

    def pre_start(self, run_control: RunControlBase) -> None:
        """Overloaded method.
        """
        file_path = Path(f"{run_control.output_file_path_base()}_data.txt")
        self.event_handler.add_custom_sink(file_path, WriteMode.TEXT, MonitorReadout.to_text,
                                           MonitorReadout.text_header(creator=self.NAME))

    def process_packet(self, packet_data: bytes) -> AbstractPacket:
        """Overloaded method.
        """
        readout = MonitorReadout.unpack(packet_data)
        seconds = readout.seconds
        self.temperature1_strip_chart.put(seconds, readout.temperature1)
        self.temperature2_strip_chart.put(seconds, readout.temperature2)
        self.humidity_strip_chart.put(seconds, readout.humidity)
        self.pressure_strip_chart.put(seconds, readout.pressure)
        self.adc1_strip_chart.put(seconds, readout.adc1)
        self.adc2_strip_chart.put(seconds, readout.adc2)
        return readout


def main() -> None:
    """Main entry point.
    """
    bootstrap_window(MonitorWindow, MonitorRunControl(), Monitor())


if __name__ == "__main__":
    main()



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for serial_.py
"""

import pytest

from baldaquin import serial_


def test_device_id(pid: int = 0x0043, vid: int = 0x2341) -> None:
    """Test the DeviceId data class.
    """
    dev_id = serial_.DeviceId(vid, pid)
    assert dev_id.vid == vid
    assert dev_id.pid == pid
    assert dev_id == serial_.DeviceId(vid, pid)
    assert dev_id != serial_.DeviceId(vid + 1, pid)
    # Check the comparison between a DeviceId object and a tuple.
    assert dev_id == (vid, pid)
    assert dev_id != (vid + 1, pid)


def test_list_com_ports() -> None:
    """Test the COM port listing.
    """
    ports = serial_.list_com_ports()
    for _ in ports:
        pass
    assert all(isinstance(port, serial_.PortInfo) for port in ports)
    ports = serial_.list_com_ports((0x2341, 0x0043))
    for _ in ports:
        pass


def test_text_line() -> None:
    """Test the simple text protocol for the serial port.
    """
    with pytest.raises(RuntimeError):
        line = serial_.TextLine.from_text("Hello world;1")
    with pytest.raises(RuntimeError):
        line = serial_.TextLine.from_text("#Hello world;1")
    with pytest.raises(RuntimeError):
        line = serial_.TextLine.from_text("Hello world;1\n")
    line = serial_.TextLine.from_text("#Hello world;1\n")
    name, version = line.unpack(str, int)
    assert name == "Hello world"
    assert version == 1
    name, version = line.unpack()
    assert name == "Hello world"
    assert version == "1"
    with pytest.raises(RuntimeError):
        name, version = line.unpack(str)
    # Test text line insertion.
    line = serial_.TextLine.from_text("#1;2;3\n")
    line.prepend("ciao")
    assert line.decode() == "#ciao;1;2;3\n"
    line.append("howdy")
    assert line.decode() == "#ciao;1;2;3;howdy\n"



# Copyright (C) 2022--2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Simplest possible mock application.
"""

from loguru import logger

from baldaquin.gui import bootstrap_window
from baldaquin.pkt import AbstractPacket
from baldaquin.silly import SILLY_APP_CONFIG
from baldaquin.silly.common import (
    SillyConfiguration,
    SillyMainWindow,
    SillyPacket,
    SillyRunControl,
    SillyUserApplicationBase,
)


class SillyRead(SillyUserApplicationBase):

    """Simplest possible user application for testing purposes.
    """

    NAME = "Silly readout"
    CONFIGURATION_CLASS = SillyConfiguration
    CONFIGURATION_FILE_PATH = SILLY_APP_CONFIG / "silly_read.cfg"

    def process_packet(self, packet_data: bytes) -> AbstractPacket:
        """Dumb data processing routine---print out the actual event.
        """
        packet = SillyPacket.unpack(packet_data)
        logger.debug(f"{packet} <- {packet_data}")
        return packet


def main() -> None:
    """Main entry point.
    """
    bootstrap_window(SillyMainWindow, SillyRunControl(), SillyRead())


if __name__ == "__main__":
    main()



# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Convenience module handling the Qt-related import.

This is mainly to handle the possibility of switching from/to PyQy/PySide
in the various available flavors. At the time the baldaquin project was started
there were basically four, slightly different sensible possibilities floating
around (PySide2/6 and PyQt5/6), each with subtle differences in semantics, PySide6
probably being the preferred choice.

Here is a (non-exhaustive) list of things that we need to keep track of, here.

The .exec() method is used in Qt to start the event loop of your QApplication or
dialog boxes. In Python 2 exec was a keyword, meaning it could not be used for
variable, function or method names. The solution used in both PyQt4 and PySide
was to rename uses of .exec to .exec_() to avoid this conflict. Python 3 removed
the exec keyword, freeing the name up to be used. As a result from PyQt6 .exec()
calls are named just as in Qt. At the time of writing, PySide6 still supports
the exec_() form of things, but PySide6 version 6.4.3 emits a deprecation warning
along the lines of '"exec_" will be removed in the future. Use "exec" instead.'
We handle this by wrapping the call into a small exec_qapp() helper function, and
we use that consistently throughout the code.

There is a rather oscure bug in matplotlib triggered by specific combinations of
matplotlib and PySide6, see
https://github.com/matplotlib/matplotlib/issues/24315
that needs some monkeypatching in matplotlib versions earlier than 3.6.2
"""

import operator
import os
import sys

from .logging_ import logger

# pylint: disable=unused-import, import-error, invalid-name

# Global flag to handle the Qt bindings in a consistent fashion.
AVAILABLE_BALDAQUIN_QT_WRAPPERS = ("PySide2", "PySide6", "PyQt5", "PyQt6")
DEFAULT_BALDAQUIN_QT_WRAPPER = "PySide6"
BALDAQUIN_QT_WRAPPER = os.getenv("BALDAQUIN_QT_WRAPPER", DEFAULT_BALDAQUIN_QT_WRAPPER)
if BALDAQUIN_QT_WRAPPER not in AVAILABLE_BALDAQUIN_QT_WRAPPERS:
    logger.error(f"{DEFAULT_BALDAQUIN_QT_WRAPPER} Qt Python wrapper is not available")
    BALDAQUIN_QT_WRAPPER = DEFAULT_BALDAQUIN_QT_WRAPPER
logger.info(f"Qt Python wrapper set to {BALDAQUIN_QT_WRAPPER}")


def _exec_qapp_old_style(qapp):
    """Old-style QApplication bootstrap call (with the final underscore).
    """
    sys.exit(qapp.exec_())


def _exec_qapp_new_style(qapp):
    """New-style QApplication bootstrap call (without the final underscore).
    """
    sys.exit(qapp.exec())


if BALDAQUIN_QT_WRAPPER == "PySide6":
    # pylint: disable=invalid-name, protected-access, no-name-in-module
    from PySide6 import QtCore, QtGui, QtWidgets
    exec_qapp = _exec_qapp_new_style
    # Horrible workaround for https://github.com/matplotlib/matplotlib/issues/24315
    from matplotlib import __version__, parse_version
    _plt_version = parse_version(__version__)
    if (_plt_version.major, _plt_version.minor, _plt_version.micro) < (3, 6, 2):
        bug_url = "https://github.com/matplotlib/matplotlib/issues/24315"
        logger.info(f"Monkeypatching matplotlib Qt compatibility layer for {bug_url}...")
        import matplotlib.backends.qt_compat as _qtcompat
        _QT_API = _qtcompat.QT_API
        _qtcompat._to_int = operator.attrgetter("value") if _QT_API in ("PyQt6", "PySide6") else int


if BALDAQUIN_QT_WRAPPER == "PySide2":
    from PySide2 import QtCore, QtGui, QtWidgets  # noqa F811
    exec_qapp = _exec_qapp_old_style


if BALDAQUIN_QT_WRAPPER == "PyQt6":
    from PyQt6 import QtCore, QtGui, QtWidgets  # noqa F811
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot
    exec_qapp = _exec_qapp_new_style


if BALDAQUIN_QT_WRAPPER == "PyQt5":
    from PyQt5 import QtCore, QtGui, QtWidgets  # noqa F811
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot
    exec_qapp = _exec_qapp_old_style



# Copyright (C) 2022--2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Binary data packet utilities.
"""

from __future__ import annotations

import struct
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum, IntEnum

from . import __version__
from .logging_ import logger
from .timeline import Timeline

DEFAULT_TEXT_PREFIX = "#"
DEFAULT_TEXT_SEPARATOR = ","


class Format(Enum):

    """Enum class encapsulating the supporte format characters from
    https://docs.python.org/3/library/struct.html#format-characters
    """

    PAD_BTYE = "x"
    CHAR = "c"
    SIGNED_CHAR = "b"
    UNSIGNED_CHAR = "B"
    BOOL = "?"
    SHORT = "h"
    UNSIGNED_SHORT = "H"
    INT = "i"
    UNSIGNED_INT = "I"
    LONG = "l"
    UNSIGNED_LONG = "L"
    LONG_LONG = "q"
    UNSIGNED_LONG_LONG = "Q"
    SSIZE_T = "n"
    SIZE_T = "N"
    FLOAT = "f"
    DOUBLE = "d"


class Layout(Enum):

    """Enum class encapsulating the supported layout characters from
    https://docs.python.org/3/library/struct.html#byte-order-size-and-alignment
    """

    NATIVE_SIZE = "@"
    NATIVE = "="
    LITTLE_ENDIAN = "<"
    BIG_ENDIAN = ">"
    NETWORK = "!"
    DEFAULT = "@"


class Edge(IntEnum):

    """Small Enum class encapsulating the edge type of a transition on a digital line.
    """

    RISING = 1
    FALLING = 0


class AbstractPacket(ABC):

    """Abstract base class for binary packets.
    """

    def __post_init__(self) -> None: # noqa: B027
        """Hook for post-initialization.
        """

    @property
    @abstractmethod
    def data(self) -> bytes:
        """Return the packet binary data.
        """

    @property
    @abstractmethod
    def fields(self) -> tuple:
        """Return the packet fields.
        """

    @abstractmethod
    def __len__(self) -> int:
        """Return the length of the binary data in bytes.
        """

    @abstractmethod
    def __iter__(self):
        """Iterate over the field values.
        """

    @abstractmethod
    def pack(self) -> bytes:
        """Pack the field values into the corresponding binary data.
        """

    @classmethod
    @abstractmethod
    def unpack(cls, data: bytes):
        """Unpack the binary data into the corresponding field values.
        """

    def _format_attributes(self, attrs: tuple[str], fmts: tuple[str] = None) -> tuple[str]:
        """Helper function to join a given set of class attributes in a properly
        formatted string.

        This is used, most notably, in the :meth:`_repr() <baldaquin.pkt.AbstractPacket._repr>`
        hook below, which in turn is used in the various ``__repr__()`` and/or ``__str__``
        implementations, and in the :meth:`to_text() <baldaquin.pkt.AbstractPacket.to_text>`
        implementations in sub-classes.

        Arguments
        ---------
        attrs : tuple
            The names of the class attributes we want to include in the representation.

        fmts : tuple, optional
            If present determines the formatting of the given attributes.
        """
        vals = (getattr(self, attr) for attr in attrs)
        if fmts is None:
            fmts = ("%s" for _ in attrs)
        return tuple(fmt % val for val, fmt in zip(vals, fmts))

    def _text(self, attrs: tuple[str], fmts: tuple[str], separator: str) -> str:
        """Helper function for text formatting.

        Note the output includes a trailing endline.

        Arguments
        ---------
        attrs : tuple
            The names of the class attributes we want to include in the representation.

        fmts : tuple,
            Determines the formatting of the given attributes.

        separator : str
            The separator between different fields.
        """
        vals = self._format_attributes(attrs, fmts)
        return f"{separator.join(vals)}\n"

    def _repr(self, attrs: tuple[str], fmts: tuple[str] = None) -> str:
        """Helper function to provide sensible string formatting for the packets.

        The basic idea is that concrete classes would use this to implement their
        `__repr__()` and/or `__str__()` special dunder methods.

        Arguments
        ---------
        attrs : tuple
            The names of the class attributes we want to include in the representation.

        fmts : tuple, optional
            If present determines the formatting of the given attributes.
        """
        vals = self._format_attributes(attrs, fmts)
        info = ", ".join([f"{attr}={val}" for attr, val in zip(attrs, vals)])
        return f"{self.__class__.__name__}({info})"

    @classmethod
    def text_header(cls, prefix: str = DEFAULT_TEXT_PREFIX, creator: str = None) -> str:
        """Hook that subclasses can overload to provide a sensible header for an
        output text file.

        Arguments
        ---------
        prefix : str
            The prefix to be prepended to each line to signal that that line is
            a comment and contains no data.

        creator : str, optional
            An optional string indicating the application that created the file.
        """
        header = f"{prefix}Created on {Timeline().latch()}\n" \
                 f"{prefix}baldaquin version: {__version__}\n"
        if creator is not None:
            header = f"{header}{prefix}Creator: {creator}\n"
        return header

    def to_text(self, separator: str = DEFAULT_TEXT_SEPARATOR) -> str:
        """Hook that subclasses can overload to provide a text representation of
        the buffer to be written in an output text file.
        """
        raise NotImplementedError


class FieldMismatchError(RuntimeError):

    """RuntimeError subclass to signal a field mismatch in a data structure.
    """

    def __init__(self, cls: type, field: str, expected: int, actual: int) -> None:
        """Constructor.
        """
        super().__init__(f'{cls.__name__} mismatch for field "{field}" '
                         f"(expected {hex(expected)}, found {hex(actual)})")


def _class_annotations(cls) -> dict:
    """Small convienience function to retrieve the class annotations.

    Note that, in order to support inheritance of @packetclasses, we do iterate
    over all the ancestors of the class at hand, starting from ``AbstractPacket``,
    and collect all the annotations along the way. The iteration is in reverse
    order, so that the final order of annotations is what one would expect.

    The try/except clause is needed because in Python 3.7 cls.__annotations__ is
    not defined when a class has no annotations, while in subsequent Python
    versions an empty dictionary is returned, instead.
    """
    # The typical mro, here, is object -> ABC -> AbstractPacket -> ...
    # Although object and ABC seem to have no annotations, and might very well
    # never have one, we only start the annotation-collection loop from
    # AbstractPacket, so that the situation is under our full control (cross your)
    # fingers.
    ancestors = cls.__mro__[:cls.__mro__.index(AbstractPacket)]
    annotations = {}
    for _cls in reversed(ancestors):
        try: # noqa: SIM105
            annotations.update(_cls.__annotations__)
        except AttributeError: #noqa: PERF203
            pass
    return annotations


def _check_format_characters(cls: type) -> None:
    """Check that all the format characters in the class annotations are valid.
    """
    for character in _class_annotations(cls).values():
        if not isinstance(character, Format):
            raise ValueError(f"Format character {character} is not a Format value")


def _check_layout_character(cls: type) -> None:
    """Check that the class layout character is valid.
    """
    cls.layout = getattr(cls, "layout", Layout.DEFAULT)
    if not isinstance(cls.layout, Layout):
        raise ValueError(f"Layout character {cls.layout} is not a Layout value")


def packetclass(cls: type) -> type:
    """Simple decorator to support automatic generation of fixed-length packet classes.
    """
    # pylint: disable = protected-access
    _check_format_characters(cls)
    _check_layout_character(cls)
    # Cache all the necessary classvariables
    annotations = _class_annotations(cls)
    cls._fields = tuple(annotations.keys())
    cls._format = f'{cls.layout.value}{"".join(char.value for char in annotations.values())}'
    cls.size = struct.calcsize(cls._format)
    # And here is a list of attributes we want to be frozen.
    cls.__frozenattrs__ = ("_fields", "_format", "size", "_data") + cls._fields

    def _init(self, *args, data: bytes = None):
        # Make sure we have the correct number of arguments---they should match
        # the class annotations.
        if len(args) != len(cls._fields):
            raise TypeError(f"{cls.__name__}.__init__() expected {len(cls._fields)} "
                            f"arguments {cls._fields}, got {len(args)}")
        # Loop over the annotations and create all the instance variables.
        for field, value in zip(cls._fields, args):
            # If a given annotation has a value attched to it, make sure we are
            # passing the same thing.
            expected = getattr(cls, field, None)
            if expected is not None and expected != value:
                raise FieldMismatchError(cls, field, expected, value)
            object.__setattr__(self, field, value)
        if data is None:
            data = self.pack()
        object.__setattr__(self, "_data", data)
        # Make sure the post-initialization is correctly performed.
        self.__post_init__()

    cls.__init__ = _init
    return cls


@packetclass
class FixedSizePacketBase(AbstractPacket):

    """Class describing a packet with fixed size.
    """

    # All of these fields will be overwritten by the @packetclass decorator, but
    # we list them here for reference, and to make pylint happy.
    _fields = None
    _format = None
    size = 0
    __frozenattrs__ = None
    _data = None

    @property
    def data(self) -> bytes:
        return self._data

    @property
    def fields(self) -> tuple:
        return self._fields

    def __len__(self) -> int:
        return self.size

    def __iter__(self):
        return (getattr(self, field) for field in self.fields)

    def pack(self) -> bytes:
        return struct.pack(self._format, *self)

    @classmethod
    def unpack(cls, data: bytes) -> AbstractPacket:
        return cls(*struct.unpack(cls._format, data), data=data)

    def __setattr__(self, key, value) -> None:
        """Overloaded method to make class instances frozen.
        """
        if key in self.__class__.__frozenattrs__:
            raise AttributeError(f"Cannot modify {self.__class__.__name__}.{key}")
        object.__setattr__(self, key, value)

    def __repr__(self):
        """String formatting.

        Note that this provides the low-level string formatting, according to the
        typical Python conventions. Subclasses are welcome to provide more
        succinct and human-readable `__str__()` special dunder implementations, but
        are encouraged not to reimplement this special method, as it inclues all
        the nitty-gritty details of the packet, which might be useful when debugging.
        """
        return self._repr(self._fields + ("data", "_format"))

    def __str__(self):
        """String formatting.

        This is a slightly more concise test representation of the packet. Again,
        subclasses can reimplement this at their leisure.
        """
        return self._repr(self._fields)

    @classmethod
    def text_header(cls, prefix: str = DEFAULT_TEXT_PREFIX, creator: str = None) -> str:
        """Overloaded method.
        """
        return f"{AbstractPacket.text_header(prefix, creator)}" \
               f'{prefix}{", ".join(cls._fields)}\n'

    def to_text(self, separator: str = DEFAULT_TEXT_SEPARATOR) -> str:
        """Overloaded method.
        """
        return f"{separator.join([str(item) for item in self])}\n"


class PacketFile:

    """Class describing a binary file containing packets.
    """

    def __init__(self, packet_class: type) -> None:
        """Constructor.
        """
        self._packet_class = packet_class
        self._input_file = None

    @contextmanager
    def open(self, file_path: str):
        """Open the file.
        """
        logger.debug(f"Opening input packet file {file_path}...")
        with open(file_path, "rb") as input_file:
            self._input_file = input_file
            yield self
            self._input_file = None
        logger.debug(f"Input file {file_path} closed.")

    def __iter__(self) -> PacketFile:
        """Return the iterator object (self).
        """
        return self

    def __next__(self) -> FixedSizePacketBase:
        """Read the next packet in the buffer.
        """
        data = self._input_file.read(self._packet_class.size)
        if not data:
            raise StopIteration
        return self._packet_class.unpack(data)

    def read_all(self) -> tuple[FixedSizePacketBase]:
        """Read in memory all the packets in the file.

        This is meant to support postprocessing applications where one needs
        all the packets in memory at the same time. Use it `cum grano salis`.
        """
        return tuple(packet for packet in self)


@dataclass
class PacketStatistics:

    """Small container class helping with the event handler bookkeeping.
    """

    packets_processed: int = 0
    packets_written: int = 0
    bytes_written: int = 0

    def reset(self) -> None:
        """Reset the statistics.
        """
        self.packets_processed = 0
        self.packets_written = 0
        self.bytes_written = 0

    def update(self, packets_processed, packets_written, bytes_written) -> None:
        """Update the event statistics.
        """
        self.packets_processed += packets_processed
        self.packets_written += packets_written
        self.bytes_written += bytes_written

    def to_dict(self) -> dict:
        """Serialization.
        """
        return self.__dict__

    @classmethod
    def from_dict(cls, **kwargs) -> PacketStatistics:
        """Deserialization.
        """
        return cls(**kwargs)



# Copyright (C) 2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for arduino_.py
"""

import contextlib

import pytest

from baldaquin import arduino_
from baldaquin.env import BALDAQUIN_SCRATCH
from baldaquin.serial_ import DeviceId

_UNO_IDS = ((0x2341, 0x0043), (0x2341, 0x0001), (0x2A03, 0x0043),
            (0x2341, 0x0243), (0x2341, 0x006A))


def test_concatenate_device_ids():
    """Test the board identiers.
    """
    assert arduino_.ArduinoBoard.concatenate_device_ids(arduino_.UNO) == _UNO_IDS


def test_board_retrieval():
    """Test the board identification code.
    """
    for device_id in _UNO_IDS:
        assert arduino_.ArduinoBoard.by_device_id(device_id) == arduino_.UNO
    assert arduino_.ArduinoBoard.by_designator("uno") == arduino_.UNO

    with pytest.raises(RuntimeError):
        arduino_.ArduinoBoard.by_device_id(DeviceId(-1, -1))

    with pytest.raises(RuntimeError):
        arduino_.ArduinoBoard.by_designator("una")


def test_autodetect():
    """Test the automatic detection of the Arduino board.
    """
    arduino_.autodetect_arduino_boards()
    arduino_.autodetect_arduino_boards(arduino_.UNO)
    arduino_.autodetect_arduino_board(arduino_.UNO)


def test_upload(test_data_path):
    """Test the sketch upload.

    Note this is within a try/except block because we cannot assume we have
    arduino-cli installed, nor a board attached to upload to.
    """
    file_path = test_data_path("blink", "blink_uno.hex")
    with contextlib.suppress(RuntimeError):
        arduino_.upload_sketch(file_path, "uno")


def test_compile(test_data_path):
    """Test the sketch compilation.

    Note this is within a try/except block because we cannot assume we have
    arduino-cli installed.
    """
    file_path = test_data_path("blink", "blink.ino")
    with contextlib.suppress(RuntimeError):
        arduino_.compile_sketch(file_path, "uno", BALDAQUIN_SCRATCH, verbose=False)


def test_project_name():
    """Test the project name machinery.
    """
    interface = arduino_.ArduinoProgrammingInterfaceBase
    for sketch_path in ("sketches/test/test.ino", "sketches/test/", "sketches/test"):
        assert interface.project_base_name(sketch_path) == "test"
        assert interface.project_name("sketches/test/test", "uno") == "test_uno"



# Copyright (C) 2022--2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Serial port interface.
"""

from __future__ import annotations

import struct
import time
from dataclasses import dataclass
from typing import Any

import serial
import serial.tools.list_ports
import serial.tools.list_ports_common

from .logging_ import logger

# List of standard baud rates.
STANDARD_BAUD_RATES = serial.Serial.BAUDRATES
DEFAULT_BAUD_RATE = 115200


@dataclass
class DeviceId:

    """Data class to hold the device id information.

    A device id is basically a tuple of two integers, the vendor id (vid) and the
    product id (pid), and the class has exactly these two members.

    Arguments
    ---------
    vid : int
        The vendor id.

    pid : int
        The product id.
    """

    vid: int
    pid: int

    def __eq__(self, other) -> bool:
        """Equality comparison.

        Note we support the comparison between a DeviceId object and a tuple.
        """
        if isinstance(other, tuple):
            return (self.vid, self.pid) == other
        return (self.vid, self.pid) == (other.vid, other.pid)

    @staticmethod
    def _hex(value: int) -> str:
        """Convenience function to format an integer as a hexadecimal string,
        gracefully handling the case when the input is not an integer.
        """
        try:
            return hex(value)
        except TypeError:
            return None

    def __hash__(self):
        """Hash function.
        """
        return hash((self.vid, self.pid))

    def __repr__(self) -> str:
        """String formatting.
        """
        return f"(vid={self._hex(self.vid)}, pid={self._hex(self.pid)})"


@dataclass
class PortInfo:

    """Small data class holding the information about a COM port.

    This is a simple wrapper around the serial.tools.list_ports_common.ListPortInfo
    isolating the basic useful functionalities, for the sake of simplicity. (And,
    in addition, this convenience class is meant to integrate the functionality
    embedded in the :class:`DeviceId` class, which is handy.)

    See https://pyserial.readthedocs.io/en/latest/tools.html#serial.tools.list_ports.ListPortInfo
    for more information.

    Arguments
    ---------
    name : str
        The name of the port (e.g., ``/dev/ttyACM0``).

    device_id : DeviceId
        The device id.

    manufacturer : str, optional
        The manufacturer of the device attached to the port.
    """

    name: str
    device_id: DeviceId
    manufacturer: str = None

    @classmethod
    def from_serial(cls, port_info: serial.tools.list_ports_common.ListPortInfo) -> PortInfo:
        """Create a Port object from a ListPortInfo object.
        """
        device_id = DeviceId(port_info.vid, port_info.pid)
        return cls(port_info.device, device_id, port_info.manufacturer)


def list_com_ports(*device_ids: DeviceId) -> list[PortInfo]:
    """List all the com ports with devices attached, possibly with a filter on
    the device ids we are interested into.

    Arguments
    ---------
    device_ids : DeviceId or vid, pid tuples, optional
        An arbitrary number of device ids to filter the list of ports returned by pyserial.
        This is useful when we are searching for a specific device attached to a
        port; an arduino uno, e.g., might look something like (0x2341, 0x43).

    Returns
    -------
    list of PortInfo objects
        The list of COM ports.
    """
    logger.info("Scanning serial devices...")
    # Populate the initial list of ports.
    ports = [PortInfo.from_serial(item) for item in serial.tools.list_ports.comports()]
    for port_info in ports:
        logger.debug(port_info)
    logger.info(f"Done, {len(ports)} device(s) found.")
    # If we're not filtering over device ids, we're done.
    if len(device_ids) == 0:
        return ports
    # Otherwise, we filter the list of ports, assuming we have any.
    if len(ports) > 0:
        # If we have a list of tuples, we convert them to DeviceId objects---this
        # will make the printout on the terminal nicer, with the 0x and all that.
        device_ids = list(device_ids)
        for i, entry in enumerate(device_ids):
            if isinstance(entry, tuple):
                device_ids[i] = DeviceId(*entry)
        logger.info(f"Filtering port list for specific devices: {device_ids}...")
        # Do the actual filtering.
        ports = [port_info for port_info in ports if port_info.device_id in device_ids]
        logger.info(f"Done, {len(ports)} device(s) remaining.")
    for port in ports:
        logger.debug(port)
    return ports


class TextLine(bytearray):

    """Small class defining a simple protocol for passing messages in string form
    over the serial port.

    A message is basically a small piece of text, in the form of a binary stream
    encoded in UTF-8, starting with a ``#`` header character and terminated by a line
    feed, and that can contain an arbitrary number of fields separated by a ``;``.
    Note the delimiters are properly checked at creation time.
    The :meth:`unpack()` class method returns a tuple with the field values,
    converted in the proper types.

    Note this is a subclass of the bytearray class, as opposed to the bytes class,
    because we provide a minimal support for prepending and appending fields to the
    text line. (Insertion at a generic index does not make much sense, as the field
    widths are variable.)

    The choice of the header character and the line feed is largely arbitrary, but
    the requirement that the line is terminated by a line feed provides the most
    straightforward way to read variable-length messages from the serial port in
    a reliable way, leveraging the serial ``readline()`` method.

    Users should by no means feel compelled to use this, but we provide it in order
    to facilitate passing strings over the serial port, saving boilerplate code
    for runtime checking.

    Example
    -------
    >>> message = TextLine.from_text("#Hello world;1\\n")
    >>> name, version = message.unpack(str, int)
    >>> print(name, type(name))
    Hello world, <class 'str'>
    >>> print(version, type(version))
    1 <class 'int'>
    """

    _ENCODING = "utf-8"
    _HEADER = "#"
    _HEADER_ORD = ord(_HEADER)
    _LINE_FEED = "\n"
    _LINE_FEED_ORD = ord(_LINE_FEED)
    _SEPARATOR = ";"

    def __init__(self, data: bytes) -> None:
        """Constructor.
        """
        super().__init__(data)
        if len(data) == 0:
            raise RuntimeError(f"Empty buffer passed to the {self.__class__.__name__} constructor")
        if not self[0] == self._HEADER_ORD:
            raise RuntimeError(f"Serial text line {self} does not start with {self._HEADER}")
        if not self[-1] == self._LINE_FEED_ORD:
            raise RuntimeError(f"Serial text line {self} does not end with a line feed")

    @classmethod
    def from_text(cls, text: str) -> bytes:
        """Create a message from a text string (mainly for debug purposes).
        """
        return cls(bytes(text, cls._ENCODING))

    def prepend(self, value: str) -> None:
        """Prepend a string field to the text line.

        Arguments
        ---------
        value : str
            The string to be prepended.
        """
        self[1:1] = bytes(f"{value}{self._SEPARATOR}", self._ENCODING)

    def append(self, value: str) -> None:
        """Append a string field to the text line.

        Arguments
        ---------
        value : str
            The string to be prepended.
        """
        self[-1:-1] = bytes(f"{self._SEPARATOR}{value}", self._ENCODING)

    def unpack(self, *converters) -> tuple:
        """Unpack the message in its (properly formatted) fields.

        Arguments
        ---------
        arguments : callable
            The converter functions to be used to unpack the fields. Note the number
            of converters passed to the function must be either zero (in which case
            all the fields are treated as strings) or equal to the number of fields
            in the message. A RuntimeError is raised if that is not the case.
        """
        # Decode the relevant part of the underlying bytes object (i.e., everything
        # but the header character and the line feed) and split the resulting string
        # by the separator character.
        fields = self[1:-1].decode(self._ENCODING).split(self._SEPARATOR)
        # Convert the fields to the proper types, if any.
        if len(converters) == 0:
            pass
        elif len(converters) == len(fields):
            fields = [converter(field) for converter, field in zip(converters, fields)]
        else:
            raise RuntimeError(f'Need exactly 0 or {len(fields)} converters to unpack "{self}"')
        # Return the fields as a tuple.
        return tuple(fields)


class SerialInterface(serial.Serial):

    """Small wrapper around the serial.Serial class.

    Note the class is designed to be instantiated without arguments, and all the
    setup is done through the :meth:`connect()` method, where the port is configured
    and the actual connection is opened.
    """

    # pylint: disable=too-many-ancestors

    def __init__(self) -> None:
        """Do-nothing constructor.
        """
        super().__init__()
        self.port_info = None

    def connect(self, port_info: PortInfo, baudrate: int = DEFAULT_BAUD_RATE,
                timeout: float = None) -> None:
        """Setup the serial connection.

        Note that the first argument is a fully-fledged PortInfo object, which
        allows to keep track of the basic information about the device connected
        to the port, in addition to configuration of the port itself. This is
        handy, e.g., for the auto-upload of arduino sketched, where we need to
        know the designation of the board we are uploading to.

        Arguments
        ---------
        port_info : PortInfo
            The PortInfo object describing the port.

        baudrate : int
            The baud rate.

            Verbatim from the pyserial documentation: the parameter baudrate can
            be one of the standard values: 50, 75, 110, 134, 150, 200, 300, 600,
            1200, 1800, 2400, 4800, 9600, 19200, 38400, 57600, 115200. These are
            well supported on all platforms.

            Standard values above 115200, such as: 230400, 460800, 500000, 576000,
            921600, 1000000, 1152000, 1500000, 2000000, 2500000, 3000000, 3500000,
            4000000 also work on many platforms and devices.

            Non-standard values are also supported on some platforms (GNU/Linux,
            MAC OSX >= Tiger, Windows). Though, even on these platforms some serial
            ports may reject non-standard values.

        timeout : float
            The timeout in seconds.

            Verbatim from the pyserial documentation: possible values for the parameter
            timeout which controls the behavior of read():

            * ``timeout = None``: wait forever / until requested number of bytes
              are received

            * ``timeout = 0``: non-blocking mode, return immediately in any case,
              returning zero or more, up to the requested number of bytes

            * ``timeout = x``: set timeout to x seconds (float allowed) returns
              immediately when the requested number of bytes are available,
              otherwise wait until the timeout expires and return all bytes that
              were received until then.
        """
        logger.debug(f"Configuring serial connection (port = {port_info.name}, "
                     f"baudarate = {baudrate}, timeout = {timeout})...")
        self.port_info = port_info
        self.port = port_info.name
        self.baudrate = baudrate
        self.timeout = timeout
        logger.info(f"Opening serial connection to port {self.port}...")
        self.open()

    def set_timeout(self, timeout: float) -> None:
        """Set the timeout for the serial port.

        Arguments
        ---------
        timeout : float
            The timeout in seconds.
        """
        logger.info(f"Setting serial connection timeout to {timeout} s...")
        self.timeout = timeout

    def disconnect(self):
        """Disconnect from the serial port.
        """
        logger.info(f"Closing serial connection to port {self.port}...")
        self.close()

    def pulse_dtr(self, pulse_length: float = 0.5) -> None:
        """Pulse the DTR line for a given amount of time.

        This asserts the DTR line, waits for a specific amount of time, and then
        de-asserts the line.

        Arguments
        ---------
        pulse_length : float
            The duration (in seconds) for the DTR line signal to be asserted.
        """
        logger.info(f"Pulsing the DTR line for {pulse_length} s...")
        self.dtr = 1
        time.sleep(pulse_length)
        self.dtr = 0

    def read_available_data(self) -> bytes:
        """Read all the available data on the serial interface.

        .. warning::
            Be cautious when using this method, as the amount of data available
            on the serial port at any given time may depend on the exact timing of
            the function call relative to what the device attached to the port is
            doing, and in the long run it might be difficult to read a stream
            of variable-length data reliably.
        """
        return self.read(self.in_waiting)

    def read_text_line(self) -> TextLine:
        """Read a line-feed terminated string data from the serial interface and return a
        :class:`TextLine` object.
        """
        return TextLine(self.readline())

    def read_and_unpack(self, fmt: str) -> Any:
        """Read a given number of bytes from the serial port and unpack them.

        Note that the number of bytes to be read from the serial port is automatically
        calculated from the format string.
        See https://docs.python.org/3/library/struct.html for all the details about
        format strings and byte ordering.

        Arguments
        ---------
        fmt : str
            The format string for the packet to be read from the seria port.

        Returns
        -------
        any
            Returns the proper Python object for the format string at hand.

        Example
        -------
        >>> s = SerialInterface(port)
        >>> val = s.read_and_unpack("B") # Single byte (val is int)
        >>> val = s.read_and_unpack(">L") # Big-endian unsigned long (val is also int)
        """
        data = self.read(struct.calcsize(fmt))
        try:
            return struct.unpack(fmt, data)[0]
        except struct.error as exception:
            logger.error(f'Could not unpack {data} with format "{fmt}".')
            raise exception

    def pack_and_write(self, value: Any, fmt: str) -> int:
        """ Pack a given value into a proper bytes object and write the latter
        to the serial port.

        Arguments
        ---------
        value : any
            The value to be written to the serial port.

        fmt : str
            The format string to pack the value with.

        Returns
        -------
        int
            The number of bytes written to the serial port.
        """
        return self.write(struct.pack(fmt, value))



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import ast
import datetime
import pathlib
import subprocess
from enum import Enum

from packaging.version import Version, parse

from baldaquin import __name__ as __package_name__

# Basic environment.
_ROOT_DIR = pathlib.Path(__file__).parent.parent
_DOCS_DIR = _ROOT_DIR / "docs"
_SRC_DIR = _ROOT_DIR / "src" / __package_name__
_VERSION_FILE_PATH = _SRC_DIR / "_version.py"
_RELEASE_NOTES_PATH = _DOCS_DIR / "release_notes.rst"
_ENCODING = "utf-8"


class BumpMode(str, Enum):

    """Small enum class describing the bump mode.
    """

    MAJOR = "major"
    MINOR = "minor"
    MICRO = "micro"


def read_version_file() -> Version:
    """Read the version string from the version file.

    Note the version file is assumed to contain exactly one line of the form
    __version__ = "version_string"
    The version string must be a valid PEP 440 version string, which is
    actively enforced by casting it to a packaging.version.Version object.
    """
    print(f"Reading version from {_VERSION_FILE_PATH}...")
    with open(_VERSION_FILE_PATH, encoding=_ENCODING) as input_file:
        line = input_file.readline().rstrip("\n")
    # If the line does not start with __version__ something went wrong...
    if not line.startswith("__version__"):
        raise ValueError(f"Invalid version file content: {line}")
    # partition() splits the string in exactly three parts around the separator.
    _, _, version_string = line.partition("=")
    # And we still need to strip the whitespaces, and get rid of the quotes.
    return parse(ast.literal_eval(version_string.strip()))


def bump_version(version: Version, mode: BumpMode) -> Version:
    """Bump the version string.
    """
    print(f'Bumping version (mode = {mode})...')
    major, minor, micro = version.release
    if mode == BumpMode.MAJOR:
        version_string = f"{major + 1}.0.0"
    elif mode == BumpMode.MINOR:
        version_string = f"{major}.{minor + 1}.0"
    elif mode == BumpMode.MICRO:
        version_string = f"{major}.{minor}.{micro + 1}"
    else:
        raise ValueError(f"Invalid bump mode {mode}")
    return Version(version_string)


def write_version_file(version: Version) -> None:
    """Write a given version string to the version file.
    """
    print(f"Writing version {version} to {_VERSION_FILE_PATH}...")
    with open(_VERSION_FILE_PATH, "w", encoding=_ENCODING) as output_file:
        output_file.write(f'__version__ = "{version}"\n')


def update_release_notes(version: Version, num_header_lines: int = 5) -> None:
    """Update the release notes file.
    """
    print(f"Updating release notes {_RELEASE_NOTES_PATH}...")
    with open(_RELEASE_NOTES_PATH, encoding=_ENCODING) as input_file:
        lines = input_file.readlines()
    text = f'Version {version} ({datetime.datetime.now().date()})'
    underline = '~' * len(text)
    text = f'\n{text}\n{underline}\n\n'
    lines.insert(num_header_lines, text)
    with open(_RELEASE_NOTES_PATH, "w", encoding=_ENCODING) as output_file:
        output_file.writelines(lines)


def _cmd(*args) -> subprocess.CompletedProcess:
    """Run a command in a subprocess.
    """
    print(f"Executing command \"{' '.join(args)}\"...")
    result = subprocess.run(args, capture_output=True, text=True, check=True)
    print(result.stdout)
    return result


def release(mode: BumpMode, target_branch: str = "main") -> None:
    """Release a new version of the package.
    """
    # Make sure we are on the proper branch---typically main.
    current_branch = _cmd("git", "branch", "--show-current").stdout.strip("\n")
    if current_branch != target_branch:
        raise RuntimeError(f"You are on the {current_branch} branch, not {target_branch}")
    # Pull the latest changes.
    _cmd("git", "pull")
    # Bump the version.
    version = bump_version(read_version_file(), mode)
    # Update the necessary files.
    write_version_file(version)
    update_release_notes(version)
    # Commit and push the modified files.
    msg = f"Prepare for tag {version}."
    _cmd("git", "commit", "-a", "-m", msg)
    _cmd("git", "push")
    msg = f"Tagging version {version}..."
    _cmd("git", "tag", "-a", str(version), "-m", f'"{msg}"')
    _cmd("git", "push", "--tags")
    _cmd("git", "status")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Release a new version of the package.")
    parser.add_argument("mode", choices=[mode.value for mode in BumpMode],
                        help="The version bump mode.")
    arguments = parser.parse_args()
    release(BumpMode(arguments.mode))



# Copyright (C) 2022--2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""System-wide facilities.
"""

from __future__ import annotations

import pathlib
import subprocess

from baldaquin._version import __version__ as __base_version__


def _git_suffix() -> str:
    """If we are in a git repo, we want to add the necessary information to the
    version string.

    This will return something along the lines of ``+gf0f18e6.dirty``.
    """
    # pylint: disable=broad-except
    kwargs = dict(cwd=pathlib.Path(__file__).parent, stderr=subprocess.DEVNULL)
    try:
        # Retrieve the git short sha to be appended to the base version string.
        args = ["git", "rev-parse", "--short", "HEAD"]
        sha = subprocess.check_output(args, **kwargs).decode().strip()
        suffix = f"+g{sha}"
        # If we have uncommitted changes, append a `.dirty` to the version suffix.
        args = ["git", "diff", "--quiet"]
        if subprocess.call(args, stdout=subprocess.DEVNULL, **kwargs) != 0:
            suffix = f"{suffix}.dirty"
        return suffix
    except (subprocess.CalledProcessError, FileNotFoundError, OSError):
        return ""


__version__ = f"{__base_version__}{_git_suffix()}"



# Copyright (C) 2022--2023 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Data buffering.
"""

from __future__ import annotations

import collections
import io
import queue
import time
from abc import ABC, abstractmethod
from collections.abc import Callable
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import Any

from .env import BALDAQUIN_ENCODING
from .logging_ import logger
from .pkt import AbstractPacket
from .profile import timing


class WriteMode(Enum):

    """Small enum class for the file write mode.

    Note this has to match the open modes in the Python open() builtin.
    """

    BINARY = "b"
    TEXT = "t"


class Sink:

    """Small class describing a file sink where a buffer can be flushed.

    Arguments
    ---------
    file_path : Path or str
        The path to the output file.

    mode : WriteMode
        The write mode (``WriteMode.BINARY`` or ``WriteMode.TEXT``)

    formatter : callable, optional
        The packet formatting function to be used to flush a buffer.

    header : anything that can be written to the output file
        The file optional file header. If not None, this gets written to the
        output file when the sink is created.
    """

    def __init__(self, file_path: Path, mode: WriteMode, formatter: Callable = None,
                 header: Any = None) -> None:
        """Constructor.
        """
        # If the output file already exists, then something has gone wrong---we
        # never overwrite data.
        if file_path.exists():
            raise FileExistsError(f"Output file {file_path} already exists")
        self.file_path = file_path
        self.formatter = formatter
        self._mode = mode
        self._output_file = None
        # Note we always open the output file in append mode.
        self._open_kwargs = dict(mode=f"a{self._mode.value}")
        if self._mode == WriteMode.TEXT:
            self._open_kwargs.update(encoding=BALDAQUIN_ENCODING)
        # At this point we do create the file and, if needed, we write the
        # header into it. (And we are ready to flush.)
        with self.open() as output_file:
            if header is not None:
                logger.debug("Writing file header...")
                output_file.write(header)

    @contextmanager
    def open(self) -> io.IOBase:
        """Open the proper file object and return it.

        Note this is implemented as a context manager, and yields a reference to
        the underlying (open) output file.
        """
        # pylint: disable=unspecified-encoding, bad-open-mode
        logger.debug(f"Opening output file {self.file_path} {self._open_kwargs}...")
        output_file = open(self.file_path, **self._open_kwargs) # noqa SIM115
        yield output_file
        output_file.close()
        logger.debug(f"Output file {self.file_path} closed.")

    def __str__(self) -> str:
        """String formatting.
        """
        if self.formatter is None:
            return f"Sink -> {self.file_path} ({self._mode})"
        return f"Sink -> {self.file_path} ({self._mode}, {self.formatter.__qualname__})"


class AbstractBuffer(ABC):

    """Abstract base class for a data buffer.

    Arguments
    ---------
    flush_size : int
        The maximum number of packets before a
        :meth:`flush_needed() <baldaquin.buf.AbstractBuffer.flush_needed()>`
        call returns True.

    flush_timeout : float
        The maximum time (in s) elapsed since the last
        :meth:`flush() <baldaquin.buf.AbstractBuffer.flush()>` before a
        :meth:`flush_needed() <baldaquin.buf.AbstractBuffer.flush_needed()>` call
        returns True.
    """

    def __init__(self, flush_size: int, flush_timeout: float) -> None:
        """Constructor.
        """
        self._flush_size = flush_size
        self._flush_timeout = flush_timeout
        self._last_flush_time = time.time()
        self._primary_sink = None
        self._custom_sinks = []

    def configure(self, flush_size: int, flush_timeout: float) -> None:
        """Configure the buffer parameters.
        """
        logger.info(f"Configuring {self.__class__.__name__} with flush_size={flush_size}, "
                    f"flush_timeout={flush_timeout} s...")
        self._flush_size = flush_size
        self._flush_timeout = flush_timeout

    def put(self, packet: AbstractPacket) -> None:
        """Put a packet into the buffer.

        Note we check in the abstract class that the packet we put into the buffer
        is an ``AbstractPacket`` instance, while the actual work is done in the
        ``_do_put()`` method, that is abstract and should be reimplemented in all
        derived classes.
        """
        if not isinstance(packet, AbstractPacket):
            raise TypeError(f"{packet} is not an AbstractPacket instance")
        self._do_put(packet)

    @abstractmethod
    def _do_put(self, packet: AbstractPacket) -> None:
        """Abstract method with the actual code to put a packet into the buffer
        (to be reimplemented in derived classes).

        .. warning::
            In case you wonder why this is called ``_do_put`` and not, e.g., ``_put()``...
            well: whoever designed the ``queue.Queue`` class (which we rely on for
            of the actual buffer implementations) apparently had the same brilliant
            idea of delegating the ``put()`` call to a ``_put()`` function, and
            overloading that name was putting things into an infinite loop.
        """

    @abstractmethod
    def pop(self) -> Any:
        """Pop an item from the buffer (to be reimplemented in derived classes).

        .. note::
            The specific semantic of `which` item is returned (e.g, the first,
            last, or something more clever) is delegated to the concrete classes,
            but we will be mostly dealing with FIFOs, i.e., unless otherwise
            stated it should be understood that this is popping items from the
            left of the queue.
        """

    @abstractmethod
    def size(self) -> int:
        """Return the number of items in the buffer (to be reimplemented in derived classes).
        """

    @abstractmethod
    def clear(self) -> None:
        """Clear the buffer (to be reimplemented in derived classes).
        """

    def almost_full(self) -> bool:
        """Return True if the buffer is almost full.
        """
        return self._flush_size is not None and self.size() >= self._flush_size

    def time_since_last_flush(self):
        """Return the time (in s) since the last flush operation, or since the
        buffer creation, in case it has never been flushed.
        """
        return time.time() - self._last_flush_time

    def flush_needed(self) -> bool:
        """Return True if the buffer needs to be flushed.
        """
        return self.almost_full() or self.time_since_last_flush() > self._flush_timeout

    def set_primary_sink(self, file_path: Path) -> Sink:
        """Set the primary sink for the buffer.
        """
        sink = Sink(file_path, WriteMode.BINARY, None, None)
        logger.info(f"Connecting buffer to primary {sink}...")
        self._primary_sink = sink
        return sink

    def add_custom_sink(self, file_path: Path, mode: WriteMode, formatter: Callable = None,
                        header: Any = None) -> Sink:
        """Add a sink to the buffer.

        See the :class:`Sink <baldaquin.buf.Sink>` class constructor for an
        explanation of the arguments.
        """
        sink = Sink(file_path, mode, formatter, header)
        logger.info(f"Connecting buffer to custom {sink}...")
        self._custom_sinks.append(sink)
        return sink

    def disconnect_sinks(self) -> None:
        """Disconnect all sinks.
        """
        self._primary_sink = None
        self._custom_sinks = []
        logger.info("All buffer sinks disconnected.")

    def _pop_and_write_raw(self, num_packets: int, output_file: io.IOBase) -> int:
        """Pop the first ``num_packets`` packets from the buffer and write the
        corresponding raw binary data into the given output file.

        Arguments
        ---------
        num_packets : int
            The number of packets to be written out.

        output_file : io.IOBase
            The output binary file.

        Returns
        -------
        int
            The number of bytes written to disk.
        """
        num_bytes_written = 0
        for _ in range(num_packets):
            num_bytes_written += output_file.write(self.pop().data)
        logger.debug(f"{num_bytes_written} bytes written to disk.")
        return num_bytes_written

    def _write(self, num_packets: int, output_file: io.IOBase, formatter: Callable) -> int:
        """Write the first ``num_packets`` packets from the buffer to the given
        output file (and with the given formatter) without popping them.

        Arguments
        ---------
        num_packets : int
            The number of packets to be written out.

        output_file : io.IOBase
            The output binary file.

        formatter : Callable
            The packet formatting function.

        Returns
        -------
        int
            The number of bytes written to disk.
        """
        num_bytes_written = 0
        for i in range(num_packets):
            num_bytes_written += output_file.write(formatter(self[i]))
        logger.debug(f"{num_bytes_written} bytes written to disk.")
        return num_bytes_written

    @timing
    def flush(self) -> tuple[int, int]:
        """Write the content of the buffer to all the sinks connected.

        .. note::
           This will write all the items in the buffer at the time of the
           function call, i.e., items added while writing to disk will need to
           wait for the next call.
        """
        if self._primary_sink is None:
            raise RuntimeError("No primary sink connected to the buffer, cannot flush")
        # Cache the number of packets to be read---this is implemented this way
        # as we might be adding new packets while flushing the buffer.
        num_packets = self.size()
        num_bytes_written = 0
        self._last_flush_time = time.time()
        # If there are no packets, then there is nothing to do, and we are not
        # actually flushing the buffer.
        if num_packets == 0:
            return (num_packets, num_bytes_written)
        # And, finally, the actual flush.
        logger.info(f"{num_packets} packets ready to be written out...")
        # First write to all the custom sinks, as this does not empty the buffer...
        for sink in self._custom_sinks:
            with sink.open() as output_file:
                num_bytes_written += self._write(num_packets, output_file, sink.formatter)
        # ... and, finally flush the buffer to the primary sink.
        with self._primary_sink.open() as output_file:
            num_raw_bytes_written = self._pop_and_write_raw(num_packets, output_file)
            num_bytes_written += num_raw_bytes_written
        logger.info(f"{num_bytes_written} bytes ({num_raw_bytes_written} raw bytes) "
                    "written to disk.")
        # Note at this point we are keeping track of both the total number of
        # bytes written to disk *and* the number of bytes written in the form of
        # raw binary packets, and we can decide which one we want to use downstream,
        # e.g., to update the GUI.
        return (num_packets, num_bytes_written)


class FIFO(queue.Queue, AbstractBuffer):

    """Implementation of a FIFO.

    This is using the `queue <https://docs.python.org/3/library/queue.html>`_
    module in the Python standard library.

    Note that the queue.Queue class is internally using a collections.deque
    object, so this is effectively another layer of complexity over the
    CircularBuffer class below. It's not entirely clear to me what the real
    difference would be, in a multi-threaded context.
    """

    def __init__(self, flush_size: int = None, flush_timeout: float = 10.) -> None:
        """Constructor.
        """
        # From the stdlib documentation: maxsize is an integer that sets the
        # upper bound limit on the number of items that can be placed in the queue.
        # If maxsize is less than or equal to zero, the queue size is infinite.
        queue.Queue.__init__(self, -1)
        AbstractBuffer.__init__(self, flush_size, flush_timeout)

    def _do_put(self, packet: AbstractPacket, block: bool = True, timeout: float = None) -> None:
        """Overloaded method.

        See https://docs.python.org/3/library/queue.html as for the meaning
        of the function arguments.
        """
        queue.Queue.put(self, packet, block, timeout)

    def pop(self) -> Any:
        """Overloaded method.
        """
        return self.get()

    def size(self) -> int:
        """Overloaded method.
        """
        return self.qsize()

    def clear(self) -> None:
        """Overloaded method.
        """
        self.queue.clear()


class CircularBuffer(collections.deque, AbstractBuffer):

    """Implementation of a simple circular buffer.

    This is a simple subclass of the Python
    `collections.deque <https://docs.python.org/3/library/collections.html#collections.deque>`_
    data structure, adding I/O facilities on top of the base class.

    Verbatim from the Python documentation: `deques support thread-safe,
    memory efficient appends and pops from either side of the deque with
    approximately the same O(1) performance in either direction.`
    For completeness, the idea of using a deque to implement a circular buffer
    comes from https://stackoverflow.com/questions/4151320
    """

    def __init__(self, flush_size: int = None, flush_timeout: float = 10.) -> None:
        """Constructor.
        """
        collections.deque.__init__(self, [], None)
        AbstractBuffer.__init__(self, flush_size, flush_timeout)

    def _do_put(self, packet: AbstractPacket) -> None:
        """Overloaded method.
        """
        self.append(packet)

    def pop(self) -> Any:
        """Overloaded method.
        """
        return self.popleft()

    def size(self) -> int:
        """Overloaded method.
        """
        return len(self)

    def clear(self) -> None:
        """Overloaded method.
        """
        collections.deque.clear(self)



# Copyright (C) 2023 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""matplotlib configuration module.
"""

import sys
from typing import Any

import matplotlib
from matplotlib import pyplot as plt

from .logging_ import logger

if sys.flags.interactive:
    plt.ion()


DEFAULT_FIG_WIDTH = 8.

DEFAULT_FIG_HEIGHT = 6.

DEFAULT_FIG_SIZE = (DEFAULT_FIG_WIDTH, DEFAULT_FIG_HEIGHT)

DEFAULT_COLORS = [
    "#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd", "#8c564b",
    "#e377c2", "#7f7f7f", "#bcbd22", "#17becf"
]


SERIF_FONTS = [
    "DejaVu Serif", "Bitstream Vera Serif", "New Century Schoolbook",
    "Century Schoolbook L", "Utopia", "ITC Bookman", "Bookman",
    "Nimbus Roman No9 L", "Times New Roman", "Times", "Palatino", "Charter",
    "serif"
]

SANS_SERIF_FONTS = [
    "DejaVu Sans", "Bitstream Vera Sans", "Lucida Grande", "Verdana", "Geneva",
    "Lucid", "Arial", "Helvetica", "Avant Garde", "sans-serif"
]

CURSIVE_FONTS = [
    "Apple Chancery", "Textile", "Zapf Chancery", "Sand", "Script MT",
    "Felipa", "cursive"
]

FANTASY_FONTS = [
    "Comic Sans MS", "Chicago", "Charcoal", "Impact", "Western", "Humor Sans",
    "xkcd", "fantasy"
]

MONOSPACE_FONTS = [
    "DejaVu Sans Mono", "Bitstream Vera Sans Mono", "Andale Mono",
    "Nimbus Mono L", "Courier New", "Courier", "Fixed", "Terminal", "monospace"
]


class PlotCard(dict):

    """Small class reperesenting a text card.

    This is essentially a dictionary that is capable of plotting itself on
    a matplotlib figure in the form of a multi-line graphic card.

    Arguments
    ---------
    data : dict
        A dictionary holding the lines to be displayed in the card.
    """

    KEY_KWARGS = dict(color="gray", size="x-small", ha="left", va="top")
    VALUE_KWARGS = dict(color="black", size="small", ha="left", va="top")

    def __init__(self, data: dict = None) -> None:
        """Constructor.
        """
        super().__init__()
        if data is not None:
            for key, value in data.items():
                self.add_line(key, value)

    def add_line(self, key: str, value: float, fmt: str = "%g", units: str = None) -> None:
        """Set the value for a given key.

        Arguments
        ---------
        key : str
            The key, i.e., the explanatory text for a given value.

        value : float, optional
            The actual value (if None, a blank line will be added).

        fmt : str
            The string format to be used to render the value.

        units : str
            The measurement units for the value.
        """
        self[key] = (value, fmt, units)

    def draw(self, axes=None, x: float = 0.05, y: float = 0.95, line_spacing: float = 0.075,
             spacing_ratio: float = 0.75) -> None:
        """Draw the card.

        Arguments
        ---------
        x0, y0 : float
            The absolute coordinates of the top-left corner of the card.

        line_spacing : float
            The line spacing in units of the total height of the current axes.

        spacing_ratio : float
            The fractional line spacing assigned to the key label.
        """
        # pylint: disable=invalid-name
        if axes is None:
            axes = plt.gca()
        key_norm = spacing_ratio / (1. + spacing_ratio)
        value_norm = 1. - key_norm
        for kwargs in (self.KEY_KWARGS, self.VALUE_KWARGS):
            kwargs["transform"] = axes.transAxes
        for key, (value, fmt, units) in self.items():
            if value is None:
                y -= 0.5 * line_spacing
                continue
            axes.text(x, y, key, **self.KEY_KWARGS)
            y -= key_norm * line_spacing
            value = fmt % value
            if units is not None:
                value = f"{value} {units}"
            axes.text(x, y, value, **self.VALUE_KWARGS)
            y -= value_norm * line_spacing


def last_line_color(default: str = "black"):
    """Return the color used to draw the last line
    """
    try:
        return plt.gca().get_lines()[-1].get_color()
    except IndexError:
        return default


def setup_axes(axes, **kwargs):
    """Setup a generic axes object.
    """
    if kwargs.get("logx"):
        axes.set_xscale("log")
    if kwargs.get("logx"):
        axes.set_yscale("log")
    xticks = kwargs.get("xticks")
    if xticks is not None:
        axes.set_xticks(xticks)
    yticks = kwargs.get("yticks")
    if yticks is not None:
        axes.set_yticks(yticks)
    xlabel = kwargs.get("xlabel")
    if xlabel is not None:
        axes.set_xlabel(xlabel)
    ylabel = kwargs.get("ylabel")
    if ylabel is not None:
        axes.set_ylabel(ylabel)
    xmin, xmax, ymin, ymax = [kwargs.get(key) for key in ("xmin", "xmax", "ymin", "ymax")]
    if xmin is None and xmax is None and ymin is None and ymax is None:
        pass
    else:
        axes.axis([xmin, xmax, ymin, ymax])
    if kwargs.get("grids"):
        axes.grid(which="both")
    if kwargs.get("legend"):
        axes.legend()


def setup_gca(**kwargs):
    """Setup the axes for the current plot.
    """
    setup_axes(plt.gca(), **kwargs)


def _set_rc_param(key: str, value: Any):
    """Set the value for a single matplotlib parameter.

    The actual command is encapsulated into a try except block because this
    is intended to work across different matplotlib versions. If a setting
    cannot be applied for whatever reason, this will happily move on.
    """
    try:
        matplotlib.rcParams[key] = value
    except KeyError:
        logger.warning(f"Unknown matplotlib rc param {key}, skipping...")
    except ValueError as exception:
        logger.warning(f"{exception}, skipping...")


def setup():
    """Basic system-wide setup.

    The vast majority of the settings are taken verbatim from the
    matplotlib 2, commit 5285e76:
    https://github.com/matplotlib/matplotlib/blob/master/matplotlibrc.template

    Note that, although this is designed to provide an experience which is as
    consistent as possible across different matplotlib versions, some of the
    functionalities are not implemented in older versions, which is why we wrap
    each parameter setting into a _set_rc_param() function call.
    """
    # pylint: disable=too-many-statements
    # http://matplotlib.org/api/artist_api.html#module-matplotlib.lines
    _set_rc_param("lines.linewidth", 1.5)
    _set_rc_param("lines.linestyle", "-")
    _set_rc_param("lines.color", DEFAULT_COLORS[0])
    # _set_rc_param("lines.marker", None)
    _set_rc_param("lines.markeredgewidth", 1.0)
    _set_rc_param("lines.markersize", 6)
    _set_rc_param("lines.dash_joinstyle", "miter")
    _set_rc_param("lines.dash_capstyle", "butt")
    _set_rc_param("lines.solid_joinstyle", "miter")
    _set_rc_param("lines.solid_capstyle", "projecting")
    _set_rc_param("lines.antialiased", True)
    _set_rc_param("lines.dashed_pattern", (2.8, 1.2))
    _set_rc_param("lines.dashdot_pattern", (4.8, 1.2, 0.8, 1.2))
    _set_rc_param("lines.dotted_pattern", (1.1, 1.1))
    _set_rc_param("lines.scale_dashes", True)

    # Markers.
    _set_rc_param("markers.fillstyle", "full")

    # http://matplotlib.org/api/artist_api.html#module-matplotlib.patches
    _set_rc_param("patch.linewidth", 1)
    _set_rc_param("patch.facecolor", DEFAULT_COLORS[0])
    _set_rc_param("patch.edgecolor", "black")
    _set_rc_param("patch.force_edgecolor", True)
    _set_rc_param("patch.antialiased", True)

    # Hatches
    _set_rc_param("hatch.color", "k")
    _set_rc_param("hatch.linewidth", 1.0)

    # Boxplot
    _set_rc_param("boxplot.notch", False)
    _set_rc_param("boxplot.vertical", True)
    _set_rc_param("boxplot.whiskers", 1.5)
    _set_rc_param("boxplot.bootstrap", None)
    _set_rc_param("boxplot.patchartist", False)
    _set_rc_param("boxplot.showmeans", False)
    _set_rc_param("boxplot.showcaps", True)
    _set_rc_param("boxplot.showbox", True)
    _set_rc_param("boxplot.showfliers", True)
    _set_rc_param("boxplot.meanline", False)
    _set_rc_param("boxplot.flierprops.color", "k")
    _set_rc_param("boxplot.flierprops.marker", "o")
    _set_rc_param("boxplot.flierprops.markerfacecolor", "none")
    _set_rc_param("boxplot.flierprops.markeredgecolor", "k")
    _set_rc_param("boxplot.flierprops.markersize", 6)
    _set_rc_param("boxplot.flierprops.linestyle", "none")
    _set_rc_param("boxplot.flierprops.linewidth", 1.0)
    _set_rc_param("boxplot.boxprops.color", "k")
    _set_rc_param("boxplot.boxprops.linewidth", 1.0)
    _set_rc_param("boxplot.boxprops.linestyle", "-")
    _set_rc_param("boxplot.whiskerprops.color", "k")
    _set_rc_param("boxplot.whiskerprops.linewidth", 1.0)
    _set_rc_param("boxplot.whiskerprops.linestyle", "-")
    _set_rc_param("boxplot.capprops.color", "k")
    _set_rc_param("boxplot.capprops.linewidth", 1.0)
    _set_rc_param("boxplot.capprops.linestyle", "-")
    _set_rc_param("boxplot.medianprops.color", DEFAULT_COLORS[1])
    _set_rc_param("boxplot.medianprops.linewidth", 1.0)
    _set_rc_param("boxplot.medianprops.linestyle", "-")
    _set_rc_param("boxplot.meanprops.color", DEFAULT_COLORS[2])
    _set_rc_param("boxplot.meanprops.marker", "^")
    _set_rc_param("boxplot.meanprops.markerfacecolor", DEFAULT_COLORS[2])
    _set_rc_param("boxplot.meanprops.markeredgecolor", DEFAULT_COLORS[2])
    _set_rc_param("boxplot.meanprops.markersize", 6)
    _set_rc_param("boxplot.meanprops.linestyle", "none")
    _set_rc_param("boxplot.meanprops.linewidth", 1.0)

    # http://matplotlib.org/api/font_manager_api.html for more
    _set_rc_param("font.family", "sans-serif")
    _set_rc_param("font.style", "normal")
    _set_rc_param("font.variant", "normal")
    _set_rc_param("font.weight", "medium")
    _set_rc_param("font.stretch", "normal")
    _set_rc_param("font.size", 14.0)
    _set_rc_param("font.serif", SERIF_FONTS)
    _set_rc_param("font.sans-serif", SANS_SERIF_FONTS)
    _set_rc_param("font.cursive", CURSIVE_FONTS)
    _set_rc_param("font.fantasy", FANTASY_FONTS)
    _set_rc_param("font.monospace", MONOSPACE_FONTS)

    # http://matplotlib.org/api/artist_api.html#module-matplotlib.text for more
    _set_rc_param("text.color", "black")

    # http://wiki.scipy.org/Cookbook/Matplotlib/UsingTex
    _set_rc_param("text.usetex", False)
    _set_rc_param("text.hinting", "auto")
    _set_rc_param("text.hinting_factor", 8)
    _set_rc_param("text.antialiased", True)
    _set_rc_param("mathtext.cal", "cursive")
    _set_rc_param("mathtext.rm", "serif")
    _set_rc_param("mathtext.tt", "monospace")
    _set_rc_param("mathtext.it", "serif:italic")
    _set_rc_param("mathtext.bf", "serif:bold")
    _set_rc_param("mathtext.sf", "sans")
    _set_rc_param("mathtext.fontset", "stixsans")
    _set_rc_param("mathtext.default", "it")

    # http://matplotlib.org/api/axes_api.html#module-matplotlib.axes
    _set_rc_param("axes.facecolor", "white")
    _set_rc_param("axes.edgecolor", "black")
    _set_rc_param("axes.linewidth", 1.25)
    _set_rc_param("axes.grid", False)
    _set_rc_param("axes.titlesize", "large")
    _set_rc_param("axes.titlepad", 6.0)
    _set_rc_param("axes.labelsize", "medium")
    _set_rc_param("axes.labelpad", 4.0)
    _set_rc_param("axes.labelweight", "normal")
    _set_rc_param("axes.labelcolor", "black")
    _set_rc_param("axes.axisbelow", "line")
    _set_rc_param("axes.formatter.limits", (-7, 7))
    _set_rc_param("axes.formatter.use_locale", False)
    _set_rc_param("axes.formatter.use_mathtext", False)
    _set_rc_param("axes.formatter.min_exponent", 0)
    _set_rc_param("axes.formatter.useoffset", True)
    _set_rc_param("axes.formatter.offset_threshold", 4)
    _set_rc_param("axes.unicode_minus", True)
    _set_rc_param("axes.autolimit_mode", "round_numbers")
    _set_rc_param("axes.xmargin", 0.)
    _set_rc_param("axes.ymargin", 0.)
    _set_rc_param("polaraxes.grid", True)
    _set_rc_param("axes3d.grid", True)

    # Dates
    _set_rc_param("date.autoformatter.year", "%Y")
    _set_rc_param("date.autoformatter.month", "%Y-%m")
    _set_rc_param("date.autoformatter.day", "%Y-%m-%d")
    _set_rc_param("date.autoformatter.hour", " %m-%d %H")
    _set_rc_param("date.autoformatter.minute", "%d %H:%M")
    _set_rc_param("date.autoformatter.second", "%H:%M:%S")
    _set_rc_param("date.autoformatter.microsecond", "%M:%S.%f")

    # see http://matplotlib.org/api/axis_api.html#matplotlib.axis.Tick
    _set_rc_param("xtick.top", False)
    _set_rc_param("xtick.bottom", True)
    _set_rc_param("xtick.major.size", 3.5)
    _set_rc_param("xtick.minor.size", 2)
    _set_rc_param("xtick.major.width", 1.25)
    _set_rc_param("xtick.minor.width", 1.)
    _set_rc_param("xtick.major.pad", 3.5)
    _set_rc_param("xtick.minor.pad", 3.4)
    _set_rc_param("xtick.color", "k")
    _set_rc_param("xtick.labelsize", "medium")
    _set_rc_param("xtick.direction", "out")
    _set_rc_param("xtick.minor.visible", False)
    _set_rc_param("xtick.major.top", True)
    _set_rc_param("xtick.major.bottom", True)
    _set_rc_param("xtick.minor.top", True)
    _set_rc_param("xtick.minor.bottom", True)
    _set_rc_param("ytick.left", True)
    _set_rc_param("ytick.right", False)
    _set_rc_param("ytick.major.size", 3.5)
    _set_rc_param("ytick.minor.size", 2)
    _set_rc_param("ytick.major.width", 1.25)
    _set_rc_param("ytick.minor.width", 1.)
    _set_rc_param("ytick.major.pad", 3.5)
    _set_rc_param("ytick.minor.pad", 3.4)
    _set_rc_param("ytick.color", "k")
    _set_rc_param("ytick.labelsize", "medium")
    _set_rc_param("ytick.direction", "out")
    _set_rc_param("ytick.minor.visible", False)
    _set_rc_param("ytick.major.left", True)
    _set_rc_param("ytick.major.right", True)
    _set_rc_param("ytick.minor.left", True)
    _set_rc_param("ytick.minor.right", True)

    # Grids
    _set_rc_param("grid.color", "#c0c0c0")
    _set_rc_param("grid.linestyle", "--")
    _set_rc_param("grid.linewidth", 0.8)
    _set_rc_param("grid.alpha", 1.0)

    # Legend
    _set_rc_param("legend.loc", "best")
    _set_rc_param("legend.frameon", True)
    _set_rc_param("legend.framealpha", 0.8)
    _set_rc_param("legend.facecolor", "inherit")
    _set_rc_param("legend.edgecolor", "gray")
    _set_rc_param("legend.fancybox", True)
    _set_rc_param("legend.shadow", False)
    _set_rc_param("legend.numpoints", 1)
    _set_rc_param("legend.scatterpoints", 1)
    _set_rc_param("legend.markerscale", 1.0)
    _set_rc_param("legend.fontsize", "medium")
    _set_rc_param("legend.borderpad", 0.4)
    _set_rc_param("legend.labelspacing", 0.5)
    _set_rc_param("legend.handlelength", 2.0)
    _set_rc_param("legend.handleheight", 0.7)
    _set_rc_param("legend.handletextpad", 0.8)
    _set_rc_param("legend.borderaxespad", 0.5)
    _set_rc_param("legend.columnspacing", 2.0)

    # See http://matplotlib.org/api/figure_api.html#matplotlib.figure.Figure
    _set_rc_param("figure.titlesize", "large")
    _set_rc_param("figure.titleweight", "normal")
    _set_rc_param("figure.figsize", DEFAULT_FIG_SIZE)
    _set_rc_param("figure.dpi", 100)
    _set_rc_param("figure.facecolor", "white")
    _set_rc_param("figure.edgecolor", "white")
    _set_rc_param("figure.autolayout", False)
    _set_rc_param("figure.max_open_warning", 20)
    _set_rc_param("figure.subplot.left", 0.125)
    _set_rc_param("figure.subplot.right", 0.95)
    _set_rc_param("figure.subplot.bottom", 0.10)
    _set_rc_param("figure.subplot.top", 0.95)
    _set_rc_param("figure.subplot.wspace", 0.2)
    _set_rc_param("figure.subplot.hspace", 0.2)

    # Images
    _set_rc_param("image.aspect", "equal")
    _set_rc_param("image.interpolation", "nearest")
    _set_rc_param("image.cmap", "jet")
    _set_rc_param("image.lut", 256)
    _set_rc_param("image.origin", "upper")
    _set_rc_param("image.resample", True)
    _set_rc_param("image.composite_image", True)

    # Contour plots
    _set_rc_param("contour.negative_linestyle", "dashed")
    _set_rc_param("contour.corner_mask", True)

    # Errorbar plots
    _set_rc_param("errorbar.capsize", 0)

    # Histogram plots
    _set_rc_param("hist.bins", 10)

    # Scatter plots
    _set_rc_param("scatter.marker", "o")

    # Saving figures
    _set_rc_param("path.simplify", True)
    _set_rc_param("path.simplify_threshold", 0.1)
    _set_rc_param("path.snap", True)
    _set_rc_param("path.sketch", None)
    _set_rc_param("savefig.dpi", "figure")
    _set_rc_param("savefig.facecolor", "white")
    _set_rc_param("savefig.edgecolor", "white")
    _set_rc_param("savefig.format", "png")
    _set_rc_param("savefig.bbox", "standard")
    _set_rc_param("savefig.pad_inches", 0.1)
    _set_rc_param("savefig.directory", "~")
    _set_rc_param("savefig.transparent", False)

    # Back-ends
    _set_rc_param("pdf.compression", 6)
    _set_rc_param("pdf.fonttype", 3)
    _set_rc_param("svg.image_inline", True)
    _set_rc_param("svg.fonttype", "path")
    _set_rc_param("svg.hashsalt", None)

    # Key maps
    _set_rc_param("keymap.fullscreen", ("f", "ctrl+f"))
    _set_rc_param("keymap.home", ("h", "r", "home"))
    _set_rc_param("keymap.back", ("left", "c", "backspace"))
    _set_rc_param("keymap.forward", ("right", "v"))
    _set_rc_param("keymap.pan", "p")
    _set_rc_param("keymap.zoom", "o")
    _set_rc_param("keymap.save", "s")
    _set_rc_param("keymap.quit", ("ctrl+w", "cmd+w"))
    _set_rc_param("keymap.grid", "g")
    _set_rc_param("keymap.grid_minor", "G")
    _set_rc_param("keymap.yscale", "l")
    _set_rc_param("keymap.xscale", ("L", "k"))

    # Animations settings
    _set_rc_param("animation.html", "none")
    _set_rc_param("animation.writer", "ffmpeg")
    _set_rc_param("animation.codec", "h264")
    _set_rc_param("animation.bitrate", -1)
    _set_rc_param("animation.frame_format", "png")
    _set_rc_param("animation.ffmpeg_path", "ffmpeg")
    _set_rc_param("animation.ffmpeg_args", "")
    _set_rc_param("animation.convert_path", "convert")

    # Miscellanea
    _set_rc_param("timezone", "UTC")
    _set_rc_param("date.converter", "concise")


setup()



# Copyright (C) 2023 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Finite-state machine diagram.
"""

from baldaquin import BALDAQUIN_DOCS
from baldaquin.plt_ import plt


def diagram_figure(figure_name):
    """Create a white, empty figure.
    """
    plt.figure(figure_name)
    plt.axis('off')


def diagram_box(x: float, y: float, text: str, **kwargs) -> None:
    """Draw a text box.
    """
    kwargs.setdefault('color', 'black')
    kwargs.setdefault('ha', 'center')
    bbox = dict(facecolor='none', edgecolor='black', boxstyle='round,pad=1')
    plt.text(x, y, text, bbox=bbox, **kwargs)


def text(x: float, y: float, text: str, **kwargs):
    """
    """
    kwargs.setdefault('ha', 'center')
    kwargs.setdefault('backgroundcolor', 'white')
    plt.text(x, y, text, **kwargs)


def arrow(start, end, connectionstyle: str = 'arc3,rad=-0.2'):
    """
    """
    _arrowprops = dict(arrowstyle="->", color="0.5", shrinkA=5, shrinkB=5,
                       patchA=None, patchB=None, connectionstyle=connectionstyle)
    plt.annotate('', xy=end, xytext=start, arrowprops=_arrowprops)


def fsm() -> None:
    """Draw the finite-state machine diagram
    """
    diagram_figure('baldaquin_fsm')
    diagram_box(0.0, 0.75, 'RESET')
    arrow((0.075, 0.8), (0.35, 0.8))
    text(0.225, 0.85, 'setup()', va='bottom')
    arrow((0.35, 0.7), (0.075, 0.7))
    text(0.225, 0.65, 'teardown()', va='top')
    diagram_box(0.45, 0.75, 'STOPPED')
    arrow((0.55, 0.8), (0.8, 0.8))
    text(0.675, 0.85, 'start_run()', va='bottom')
    arrow((0.8, 0.7), (0.55, 0.7))
    text(0.675, 0.65, 'stop_run()', va='top')
    diagram_box(0.9, 0.75, 'RUNNING')
    diagram_box(0.675, 0.25, 'PAUSED')
    arrow((0.95, 0.7), (0.775, 0.25))
    text(0.875, 0.4, 'pause()')
    arrow((0.725, 0.325), (0.9, 0.7))
    text(0.775, 0.5, 'resume()')
    arrow((0.575, 0.25), (0.45, 0.7))
    text(0.475, 0.45, 'stop()')


if __name__ == '__main__':
    fsm()
    file_path = BALDAQUIN_DOCS / 'figures' / 'baldaquin_fsm.png'
    plt.savefig(file_path)
    plt.show()



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Logging facilities.
"""

import sys

from loguru import logger

from . import __name__ as __package_name__

# Initial logger configuration: disable all logging removing all sinks.
logger.remove()
# Disable log emission from this namespace by default.
logger.disable(__package_name__)


def setup_logger(level: str = "DEBUG") -> int:
    """Minimal configuration facility for the loguru logger.

    A few remarks about the loguru internals. In order to keep the API clean, the
    author of the library made the deliberate decision not to allow to change handlers,
    so that the preferred way to change the logger configuration is to remove all the
    existing handlers and start from scratch---this is exactly what we are doing here.

    Also note that whenever you add a new handler, you get back an ID that can be used
    to remove the handler later on. The default handler (which we get rid of at the
    first call to this function) is guaranteed to have ID 0.

    Arguments
    ---------
    level : str
        The minimum logging level to be used by the logger. Defaults to "DEBUG".
        Other possible values are "INFO", "WARNING", "ERROR", and "CRITICAL".

    Returns
    -------
    int
        The ID of the handler that was added.
    """
    # Enable logging for this namespace.
    logger.enable(__package_name__)
    # Remove all existing handlers.
    logger.remove()
    # Create a plain, terminal-based logger.
    fmt = ">>> <level>[{level}] {message}</level>"
    return logger.add(sys.stderr, level=level, colorize=True, format=fmt)


def start_file_logging(file_path: str, level: str = "DEBUG") -> int:
    """Add a new file-based handler to the logger.

    Arguments
    ---------
    file_path : str
        The path to the file where the log messages should be written.

    level : str
        The minimum logging level to be used by the logger. Defaults to "DEBUG".

    Returns
    -------
    int
        The ID of the handler that was added.
    """
    logger.info(f"Directing logger output to {file_path} (level={level})...")
    return logger.add(file_path, level=level)



# Copyright (C) 2022--2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Configuration facilities.
"""

import datetime
import json
from pathlib import Path
from typing import Any

from .env import BALDAQUIN_DATA, BALDAQUIN_ENCODING
from .logging_ import logger


class ConfigurationParameter:

    """Class representing a configuration parameter.

    This is a simple attempt at putting in place a generic configuration mechanism
    where we have some control on the values we are passing along.

    A configuration parameter is fully specified by its name, type and value, and
    When setting the latter, we make sure that the its type matches.
    Additional, we can specify simple conditions on the parameters that are
    then enforced at runtime.

    Arguments
    ---------
    name : str
        The parameter name.

    type_ : type
        The parameter type.

    value : anything
        The parameter value.

    intent : str
        The intent of the parameter, acting as a comment in the corresponding
        configuration file.

    units : str, optional
        The units for the configuration parameter.

    fmt : str, optional
        An optional format string for the preferred rendering the parameter value.

    constraints : dict, optional
        A dictionary containing optional specifications on the parameter value.
    """

    # Do not remove the following comments, they are used by sphinx to generate the
    # documentation, see https://stackoverflow.com/questions/31561895
    # Start definition of valid constraints.
    _VALID_CONSTRAINTS = {
        int: ("choices", "step", "min", "max"),
        float: ("min", "max"),
        str: ("choices",)
    }
    # End definition of valid constraints.

    def __init__(self, name: str, type_: type, value: Any, intent: str,
                 units: str = None, fmt: None = None, **constraints) -> None:
        """Constructor.
        """
        self.name = name
        self.type = type_
        self.value = None
        self.intent = intent
        self.units = units
        self.fmt = fmt
        for key in tuple(constraints):
            if key not in self._VALID_CONSTRAINTS.get(self.type, ()):
                raise RuntimeError(f"Invalid spec ({key}) for {self.name} ({self.type})")
        self.constraints = constraints
        self.set_value(value)

    def _check_range(self, value: Any) -> None:
        """Generic function to check that a given value is within a specified range.
        """
        if "min" in self.constraints and value < self.constraints["min"]:
            raise RuntimeError(f"Value {value} is too small for {self}")
        if "max" in self.constraints and value > self.constraints["max"]:
            raise RuntimeError(f"Value {value} is too large for {self}")

    def _check_choices(self, value: Any) -> None:
        """Generic function to check that a parameter value is within the
        allowed choices.
        """
        if "choices" in self.constraints and value not in self.constraints["choices"]:
            raise RuntimeError(f"Unexpected choice {value} for {self}")

    def _check_step(self, value: int) -> None:
        """Generic function to check the step size for an integer.
        """
        delta = value - self.constraints.get("min", 0)
        if "step" in self.constraints and delta % self.constraints["step"] != 0:
            raise RuntimeError(f"Invalid value {value} for {self}")

    def set_value(self, value: Any) -> None:
        """Set the parameter value.

        Note that this is where all the runtime checking is performed.
        """
        # Make sure that the value we are passing is of the right type.
        if not isinstance(value, self.type):
            raise RuntimeError(f"Invalid type {value} ({type(value).__name__}) for {self}")
        # Make all the necessary checks on the constraints, if necessary.
        if self.constraints:
            if self.type is int:
                self._check_choices(value)
                self._check_range(value)
                self._check_step(value)
            elif self.type is float:
                self._check_range(value)
            elif self.type is str:
                self._check_choices(value)
        # And if we made it all the way to this point we're good to go :-)
        self.value = value

    def formatted_value(self) -> str:
        """Return the formatted parameter value (as a string).
        """
        if self.fmt is None:
            return f"{self.value}"
        return f"{self.value:{self.fmt}}"

    def pretty_print(self) -> str:
        """Return a pretty-printed string representation of the parameter.
        """
        text = f"{self.name:.<30} {self.formatted_value()}"
        if self.units:
            text += f" {self.units}"
        return text

    def __str__(self):
        """String formatting.
        """
        return f'{self.__class__.__name__} "{self.name}" ({self.type.__name__}, {self.constraints})'


class ConfigurationSectionBase(dict):

    """Base class for a single section of a configuration object. (Note this class
    is not meant to be instantiated, but rather subclassed, overriding the proper
    class variables as explained below.)

    The basic idea, here, is that specific configuration classes simply override
    the ``TITLE`` and ``_PARAMETER_SPECS`` class members, the latter encoding
    the name, type and default values for all the configuration parameters, as well
    as optional help strings and constraints.

    The class interface is fairly minimal, with support to set, and retrieve
    parameter values, and for string formatting.
    """

    TITLE = None
    _PARAMETER_SPECS = ()

    def __init__(self) -> None:
        """Constructor.
        """
        super().__init__()
        # Parse the parameters specs. The basic rule, here, is that if the last
        # element in the spec tuple is a dictionary, we are interpreting it the
        # map of constraints for the parameter, which is otherwise empty.
        # Everything else represent the constructor arguments, in order.
        for specs in self._PARAMETER_SPECS:
            if isinstance(specs[-1], dict):
                args, constraints = specs[:-1], specs[-1]
            else:
                args, constraints = specs, {}
            parameter = ConfigurationParameter(*args, **constraints)
            self[parameter.name] = parameter

    def set_value(self, parameter_name, value) -> None:
        """Update the value of a configuration parameter.
        """
        try:
            # Note all the runtime checking happens in ConfigurationParameter.set_value()
            self[parameter_name].set_value(value)
        except KeyError as exc:
            raise RuntimeError(f'Unknown parameter "{parameter_name}" for '
                               f'{self.__class__.__name__} "{self.TITLE}"') from exc

    def value(self, parameter_name) -> Any:
        """Return the value for a given parameter.
        """
        return self[parameter_name].value

    def formatted_value(self, parameter_name) -> str:
        """Return the formatted value for a given parameter.
        """
        return self[parameter_name].formatted_value()

    def as_dict(self):
        """Return a view on the configuration in the form of a {name: value} dictionary
        representing the underlying configuration parameters.

        This is used downstream to serialize the configuration and writing it to
        file.
        """
        return {parameter.name: parameter.value for parameter in self.values()}

    def __str__(self) -> str:
        """String formatting.
        """
        data = "".join(f"{param.pretty_print()}\n" for param in self.values())
        return f"---------------{self.TITLE:-<25}\n{data}"


class LoggingConfigurationSection(ConfigurationSectionBase):

    """Configuration section for the logging.
    """

    TITLE = "Logging"
    _LOGGING_LEVELS = ("TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL")
    _PARAMETER_SPECS = (
        ("terminal_level", str, "DEBUG", "Terminal logging level", dict(choices=_LOGGING_LEVELS)),
        ("file_level", str, "DEBUG", "File logging level", dict(choices=_LOGGING_LEVELS))
    )


class BufferingConfigurationSection(ConfigurationSectionBase):

    """Configuration section for the packet buffering.
    """

    TITLE = "Buffering"
    _PARAMETER_SPECS = (
        ("flush_size", int, 100, "Flush size", dict(min=1)),
        ("flush_timeout", float, 10., "Flush timeout", "s", ".3f", dict(min=1.))
    )


class MulticastConfigurationSection(ConfigurationSectionBase):

    """Configuration section for the packet multicasting.
    """

    TITLE = "Multicast"
    _PARAMETER_SPECS = (
        ("enabled", bool, False, "Enable multicast"),
        ("ip_address", str, "127.0.0.1", "IP address"),
        ("port", int, 20004, "Port", dict(min=1024, max=65535))
    )


class Configuration(dict):

    """Class describing a configuration object, that is, a dictionary of
    instances of ConfigurationSectionBase subclasses.

    Configuration objects provide file I/O through the JSON protocol. One
    important notion, here, is that configuration objects are always created
    in place with all the parameters set to their default values, and then updated
    from a configuration file. This ensures that the configuration is always
    valid, and provides an effective mechanism to be robust against updates of
    the configuration structure.
    """

    def __init__(self, *sections: ConfigurationSectionBase) -> None:
        """Constructor.
        """
        super().__init__()
        for section in sections:
            self.add_section(section)

    def add_section(self, section: ConfigurationSectionBase) -> None:
        """Add a section to the configuration.
        """
        self[section.TITLE] = section

    def update_from_file(self, file_path: str) -> None:
        """Update the configuration dictionary from a JSON file.

        Note we try and catch here all the possible exceptions while updating a
        file, and if anything happens during the update we create a timestamped
        copy of the original file so that the thing can be debugged at later time.
        The contract is that the update always proceeds to the end, and all the
        fields that can be legitimately updated get indeed updated.
        """
        logger.info(f"Updating configuration from {file_path}...")
        with open(file_path, encoding=BALDAQUIN_ENCODING) as input_file:
            data = json.load(input_file)
        errors = False
        for title, section_data in data.items():
            try:
                section = self[title]
            except KeyError:
                logger.warning(f'Unknown configuration section "{title}"')
                errors = True
                continue
            for parameter_name, value in section_data.items():
                try:
                    section.set_value(parameter_name, value)
                except RuntimeError as exc: # noqa: PERF203
                    logger.warning(exc)
                    errors = True
        if errors:
            timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S.%f")
            file_name = f"{Path(file_path).name}.backup.{timestamp}"
            dest = BALDAQUIN_DATA / file_name
            logger.warning(f"Error(s) during update, copying original file to {dest}...")

    def as_dict(self) -> dict:
        """Return a view on the configuration in the form of a dictionary that
        can be used for serialization.
        """
        return {title: section.as_dict() for title, section in self.items()}

    def to_json(self, indent: int = 4) -> str:
        """Encode the configuration into JSON to be written to file.
        """
        return json.dumps(self.as_dict(), indent=indent)

    def save(self, file_path: str) -> None:
        """Dump the configuration dictionary to a JSON file.
        """
        logger.info(f"Writing configuration dictionary to {file_path}...")
        with open(file_path, "w", encoding=BALDAQUIN_ENCODING) as output_file:
            output_file.write(self.to_json())

    def __str__(self):
        """String formatting.
        """
        text = "".join(f"{section}" for section in self.values())
        return text.strip("\n")


class UserApplicationConfiguration(Configuration):

    """Base class for a generic user application configuration.
    """

    _USER_APPLICATION_SECTION_TITLE = "User Application"
    _PARAMETER_SPECS = ()

    def __init__(self) -> None:
        """Constructor.
        """
        super().__init__()
        self.add_section(LoggingConfigurationSection())
        self.add_section(BufferingConfigurationSection())
        self.add_section(MulticastConfigurationSection())

        class _UserApplicationConfigurationSection(ConfigurationSectionBase):

            TITLE = self._USER_APPLICATION_SECTION_TITLE
            _PARAMETER_SPECS = self._PARAMETER_SPECS

        self.add_section(_UserApplicationConfigurationSection())

    def overwrite_section(self, section: ConfigurationSectionBase) -> None:
        """Overwrite a section in the configuration.
        """
        if section.TITLE not in self:
            raise RuntimeError(f'Unknown section "{section.TITLE}" for '
                               f'{self.__class__.__name__}')
        self[section.TITLE] = section

    def logging_section(self):
        """Return the logging section of the configuration.
        """
        return self[LoggingConfigurationSection.TITLE]

    def buffering_section(self):
        """Return the buffering section of the configuration.
        """
        return self[BufferingConfigurationSection.TITLE]

    def multicast_section(self):
        """Return the multicast section of the configuration.
        """
        return self[MulticastConfigurationSection.TITLE]

    def application_section(self):
        """Return the user application section of the configuration.
        """
        return self[self._USER_APPLICATION_SECTION_TITLE]
